from functools import partial
from typing import Any, Dict, List, Optional

import click

from gcs_cli.helpers import requires_login
from gcs_cli.parsing import globus_command
from gcs_cli.prettyprint import FORMAT_TEXT_TABLE, formatted_print
from gcs_cli.sdk_client.auth_client import MissingConsent
from gcs_cli.services.auth import get_client


def format_domains(key: str, policy: Dict[str, Any]) -> str:
    domains: Optional[List[str]] = policy[f"domain_constraints_{key}"]
    if not domains:
        return "None"
    if len(domains) == 1:
        return domains[0]
    return domains[0] + ", ..."


@globus_command(
    "list",
    short_help="List all authentication policies available to you",
    disable_options=["explicit_host"],
)
@requires_login
def auth_policy_list() -> None:
    """
    List all authentication policies on all Auth projects for which the current
    user is an admin.
    """
    client = get_client()
    try:
        res = client.get_auth_policy_list()
    except MissingConsent as e:
        raise click.ClickException(
            "Missing consent for the policy API. Please run the following command:\n"
            f"    globus-connect-server session consent {e.scope}"
        )

    formatted_print(
        res,
        text_format=FORMAT_TEXT_TABLE,
        fields=[
            ("Display Name", "display_name"),
            ("ID", "id"),
            ("Include Domains", partial(format_domains, "include")),
            ("Exclude Domains", partial(format_domains, "exclude")),
            ("High Assurance", "high_assurance"),
        ],
    )
