from typing import Any, Optional

import click

from gcs_cli.helpers import deny_client_login

import globus_sdk
from globus_sdk.scopes import AuthScopes
from globus_sdk import AuthAPIError

from gcs_cli.config import (
    AUTH_CLIENT_DATA_CONFIG_KEY,
    GCS_CLIENT_DATA_CONFIG_KEY,
    internal_auth_client,
    internal_native_client,
    load_client_data,
    token_storage_adapter,
)
from gcs_cli.parsing import globus_command

_RESCIND_HELP = """\
Rescinding Consents
-------------------
The logout command only revokes tokens that it can see in its storage.
If you are concerned that logout may have failed to revoke a token,
you may want to manually rescind the GCS CLI consent on the
Manage Consents Page:

    https://auth.globus.org/consents
"""


_LOGOUT_EPILOG = """\
You are now successfully logged out of the GCS CLI.
You may also want to logout of any browser session you have with Globus:

  https://auth.globus.org/v2/web/logout

Before attempting any further CLI commands, you will have to login again using

  globus-connect-server login ENDPOINT_ID
"""


@globus_command(
    "logout",
    short_help="Logout of the GCS CLI",
    disable_options=["format", "explicit_host"],
)
@click.confirmation_option(
    prompt="Are you sure you want to logout?",
    help='Automatically say "yes" to all prompts',
)
@deny_client_login("Clients are automatically 'logged out' after each command")
def logout_command() -> None:
    """
    Logout of the GCS CLI. Removes your Globus tokens from local storage, and revokes
    them so that they cannot be used anymore.
    """
    # try to get the user's preferred username from userinfo
    # if an API error is raised, they probably are not logged in

    adapter = token_storage_adapter()
    gcs_client_data = adapter.read_config(GCS_CLIENT_DATA_CONFIG_KEY)
    gcs_client_id: Optional[str] = None
    gcs_fqdn: Optional[str] = None
    if gcs_client_data is not None:
        gcs_client_id = gcs_client_data.get("client_id")
        gcs_fqdn = gcs_client_data.get("fqdn")

    if (not gcs_client_id) or (not gcs_fqdn):
        click.echo(
            (
                "Unable to lookup GCS information. You may not be logged in. "
                "Attempting logout anyway...\n"
            )
        )
    else:
        click.echo(
            "Logging out of GCS endpoint {} at {}\n".format(gcs_client_id, gcs_fqdn)
        )

    # we use a Native Client to prevent an invalid instance client
    # from preventing token revocation
    native_client = internal_native_client()

    # Find GCS tokens
    token_data: list[dict[str, Any]] = []
    if gcs_client_id:
        gcs_token_data = adapter.get_token_data(gcs_client_id) or {}
        if gcs_token_data:
            token_data.append(gcs_token_data)

    # Find Auth tokens
    auth_token_data = adapter.get_token_data(AuthScopes.resource_server) or {}
    if auth_token_data:
        token_data.append(auth_token_data)

    # remove tokens from config and revoke them
    # also, track whether or not we should print the rescind help
    print_rescind_help = False
    for td in token_data:
        for token_opt in ("access_token", "refresh_token"):
            # first lookup the token -- if not found we'll continue
            token = td.get(token_opt)
            if not token:
                print_rescind_help = True
                continue
            # token was found, so try to revoke it
            try:
                native_client.oauth2_revoke_token(token)
            # if we network error, revocation failed -- print message and abort so
            # that we can revoke later when the network is working
            except globus_sdk.NetworkError:
                click.echo(
                    (
                        "Failed to reach Globus to revoke tokens. "
                        "Because we cannot revoke these tokens, canceling "
                        "logout"
                    )
                )
                click.get_current_context().exit(1)

        adapter.remove_tokens_for_resource_server(td["resource_server"])

    # delete the instance client if one exists
    client_data = load_client_data()

    if client_data:
        instance_client = internal_auth_client()
        try:
            instance_client.delete(
                "/v2/api/clients/{}".format(client_data["client_id"])
            )
        # if the client secret has been invalidated or the client has
        # already been removed, we continue on
        except AuthAPIError:
            pass

    # remove deleted client values
    adapter.remove_config(AUTH_CLIENT_DATA_CONFIG_KEY)
    adapter.remove_config(GCS_CLIENT_DATA_CONFIG_KEY)

    # if print_rescind_help is true, we printed warnings above
    # so, jam out an extra newline as a separator
    click.echo(("\n" if print_rescind_help else "") + _LOGOUT_EPILOG)

    # if some token wasn't found in the config, it means its possible that the
    # config file was removed without logout
    # in that case, the user should rescind the CLI consent to invalidate any
    # potentially leaked refresh tokens, so print the help on that
    if print_rescind_help:
        click.echo(_RESCIND_HELP)
