from typing import (
    Any,
    Dict,
    Generator,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
)
import ipaddress
import enum

import click

from globus_sdk.paging import Paginator

from gcs_cli.helpers import requires_login
from gcs_cli.parsing import globus_command
from gcs_cli.prettyprint import FORMAT_TEXT_TABLE, formatted_print
from gcs_cli.services.gcs import get_client


def format_ip_addresses(node: Dict[str, Any]) -> str:
    ip_addresses: List[str] = node["ip_addresses"]
    first = ip_addresses[0]
    if len(ip_addresses) == 1:
        return first
    else:
        return first + ", ..."


def _ip_sort_key(
    ip: str,
) -> Tuple[int, Union[ipaddress.IPv4Address, ipaddress.IPv6Address]]:
    """
    Key function for sorting IPv4 and IPv6 address strings. Returns a tuple of IP
    version and an ipaddress object.
    """
    addr = ipaddress.ip_address(ip)
    return (addr.version, addr)


def _sort_nodes_by_ip(
    data: Iterable[Mapping[str, Any]]
) -> Generator[Mapping[str, Any], None, None]:
    """
    Returns a Generator which returns each node record in order by IP. IPs within
    each node record are ordered as well.

    :param data:
        Iterable of node data records (DATA_TYPE: node#x.y.z)

    This function prefers IPv4 over IPv6. Nodes are first sorted by their own, possibly
    multiple, ip addresses. Then the node return order is sorted by each node's first
    sorted IP address. IPv4 nodes are returned first, then IPv6 nodes.

    Only supported from a Text output context; JSON output contexts return a different
    DATA_TYPE.
    """
    nodes = [node for node in data]

    for node in nodes:
        node["ip_addresses"].sort(key=_ip_sort_key)

    nodes.sort(key=lambda node: _ip_sort_key(node["ip_addresses"][0]))
    for node in nodes:
        yield node


class _SortField(enum.Enum):
    IP = enum.auto()


@globus_command("list", short_help="List all Nodes on the Endpoint")
@click.option(
    "--sort",
    metavar="FIELD",
    type=click.Choice([x.name for x in _SortField], case_sensitive=False),
    help=(
        "Sort table entries in TEXT format. "
        f"Supported values for FIELD: {', '.join([x.name for x in _SortField])}."
    ),
)
@requires_login
def node_list(sort: Optional[str]) -> None:
    """
    Show all Nodes on the current Endpoint

    If a node has multiple IPs, only one will be shown in text output. Use JSON output
    or `node show` to see all IPs.
    """
    client = get_client()
    res: Paginator = client.paginated.get_node_list()
    formatted_print(
        res,
        text_format=FORMAT_TEXT_TABLE,
        fields=[
            ("ID", "id"),
            ("IP Addresses", format_ip_addresses),
            ("Status", "status"),
        ],
        sort_callback=_sort_nodes_by_ip if sort == _SortField.IP.name else None,
    )
