import time
from typing import Any, Dict, cast

import globus_sdk

from gcs_cli.config import internal_auth_client, token_storage_adapter
from gcs_cli.helpers import deny_client_login, print_command_hint, requires_login
from gcs_cli.parsing import globus_command
from gcs_cli.prettyprint import formatted_print
from gcs_cli.services.auth import get_client


def _is_mfa(authentication: Dict[str, Any]) -> bool:
    """
    Local helper function that determins if session authentication was
    performed using multi-factor authentication.
    """
    if "amr" not in authentication:
        return False
    if not isinstance(authentication["amr"], list):
        return False
    return "mfa" in authentication["amr"]


@globus_command(
    "show",
    short_help="Show your current auth session",
    disable_options=["explicit_host"],
)
@deny_client_login("Clients are always in session.")
@requires_login
def session_show() -> None:
    """List all identities in your current CLI auth session."""
    # get a token to introspect, refreshing if neccecary
    user_client = get_client()
    if user_client.authorizer:
        # forces a token refresh if needed
        user_client.authorizer.get_authorization_header()

    access_token = (
        token_storage_adapter().get_token_data("auth.globus.org") or {}
    ).get("access_token")
    assert access_token

    app_client = internal_auth_client()
    res = app_client.oauth2_token_introspect(access_token, include="session_info")

    session_info = cast(Dict, res.get("session_info", {}))
    authentications = session_info.get("authentications") or {}

    # resolve ids to human readable usernames
    resolved_ids = globus_sdk.IdentityMap(user_client, list(authentications))

    # put the nested dicts in a format table output can work with
    # while also converting vals into human readable formats
    list_data = [
        {
            "id": key,
            "username": resolved_ids.get(key, {}).get("username"),
            "auth_time": time.strftime(
                "%Y-%m-%d %H:%M %Z", time.localtime(vals["auth_time"])
            ),
            "mfa": _is_mfa(vals),
        }
        for key, vals in authentications.items()
    ]

    print_command_hint(
        "For information on your primary identity or full identity set see\n"
        "  globus-connect-server whoami\n"
    )

    formatted_print(
        list_data,
        json_converter=lambda x: session_info,
        fields=[
            ("Username", "username"),
            ("ID", "id"),
            ("Auth Time", "auth_time"),
            ("MFA", "mfa"),
        ],
    )
