from typing import TYPE_CHECKING, Any, Dict, Tuple

import click

if TYPE_CHECKING:
    import uuid

from gcs_cli.commands.storage_gateway._common import make_update_command
from gcs_cli.parsing import (
    box_options,
    globus_command,
    storage_gateway_create_and_update_params,
    storage_gateway_id_arg,
)
from gcs_cli.sdk_client.data import BoxStoragePolicies, StorageGatewayDocument
from gcs_cli.services.gcs import BOX_CONNECTOR_ID


@globus_command("box", short_help="Update a Box Storage Gateway")
@storage_gateway_id_arg
@storage_gateway_create_and_update_params(create=False)
@box_options
@make_update_command
def box_update(
    storage_gateway_id: "uuid.UUID",
    *,
    box_settings: Dict[str, Any],
    box_client_id: str,
    box_client_secret: str,
    box_user_api_rate_limit: int,
    **kwargs: Any
) -> Tuple["uuid.UUID", StorageGatewayDocument]:
    """
    Update an existing Box Storage Gateway.
    Any options given will replace existing values in the storage gateway's definition.
    Any options not given will retain their existing values.
    """
    if box_settings and (box_client_id or box_client_secret):
        raise click.UsageError(
            "Unable to use --box-settings and --box-client-id or --box-client-secret"
        )
    policies = None
    args: Dict[str, Any] = {}
    if box_client_id is not None:
        args["client_id"] = box_client_id
    if box_client_secret is not None:
        args["secret"] = box_client_secret
    if box_settings:
        args["enterpriseID"] = box_settings["enterpriseID"]
        args["boxAppSettings"] = box_settings["boxAppSettings"]
    if box_user_api_rate_limit is not None:
        args["user_api_rate_limit"] = box_user_api_rate_limit

    if args:
        policies = BoxStoragePolicies(**args)

    return (
        storage_gateway_id,
        StorageGatewayDocument(
            connector_id=BOX_CONNECTOR_ID, policies=policies, **kwargs
        ),
    )
