import enum
import typing as t
import uuid

from flask import (
    current_app,
    g,
    jsonify,
)

from globus.manager.api.blueprints.decorators import (
    authenticated,
    enum_list_converter,
    expects_json,
    query_parameter,
    requires_endpoint_visibility,
    requires_one_role,
    result_data,
    returns_error,
)
from globus.manager.api.blueprints.utils import update_object
from globus.manager.api.blueprints.decorators import (
    Base62Marker,
    paginated,
    Paginator,
)
import globus.manager.api.models as m
from globus.manager.api.models.result import Result
import globus.manager.api.errors as errors
from globus.manager.api.schemas.api.result import result_schema
from globus.manager.api.schemas.api.node import (
    node_patch_schema,
    node_schema,
)
from .openapi import OpenAPIBlueprint


blueprint = OpenAPIBlueprint(
    "nodes",
    __name__,
    tag="Nodes",
    description="""
    Services for GCS endpoints may be deployed on multiple different physical
    resources.  The Nodes API allows an administrator to manage those nodes.

    """,
    menu_weight=5,
    url_prefix="/api/",
)


@blueprint.route(  # type: ignore
    "/nodes",
    methods=["GET"],
    summary="List endpoint nodes",
    operationId="listNodes",
)
@result_data(
    description="List nodes response",
    schemas=[node_schema],
)
@authenticated
@paginated
@requires_endpoint_visibility
def nodes_get(*, page_size: int, marker: t.Optional[Paginator]) -> Result:
    """
    Get the endpoint's list of nodes.

    This operation requires either the endpoint to have the `public` property
    set to true or the caller to have a role that allows viewing this endpoint
    or a collection on it.
    """
    wildcard_domain = m.GlobusDnsDomain.query.one_or_none()
    if not wildcard_domain:
        raise errors.InternalServerError("Server WILDCARD_DOMAIN is not configured")

    nodes = sorted(wildcard_domain.nodes, key=lambda x: x.id)

    current_app.logger.debug("Marker is %s", marker.marker if marker else None)
    if marker:
        nodes = [n for n in nodes if n.id >= marker.marker]

    if len(nodes) > page_size:
        next_page_token = Base62Marker(marker_data=nodes[page_size].id)
        nodes = nodes[:page_size]
    else:
        next_page_token = None

    res = result_schema.dump(
        Result(data=[node_schema.dump(n) for n in nodes], marker=next_page_token)
    )

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route(
    "/nodes",
    methods=["POST"],
    summary="Create a new node",
    operationId="postNode",
)
@result_data(
    description="Create node response",
    schemas=[node_schema],
    max_items=1,
)
@returns_error(error_type=errors.ForbiddenError, error_detail=[errors.LimitExceeded])
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
@expects_json(
    schema=node_schema,
    description="New node definition",
    required=True,
)
def nodes_post() -> Result:
    """
    Create a new node to describe a host which is providing service
    for this endpoint. This adds the node's IP address to the DNS record
    for this endpoint's GCS Manager and for all collections.

    On success returns a copy of the created Node with the system generated id
    added.
    """
    new_node = g.validated_data
    if "id" in new_node:
        del new_node["id"]
    new_node["id"] = str(uuid.uuid4())

    with current_app.appsync_client.pause_notifications():
        wildcard_domain = m.GlobusDnsDomain.query.one_or_none()
        if not wildcard_domain:
            raise errors.InternalServerError("Server is misconfigured")

        _check_for_node_conflicts(g.validated_data, wildcard_domain, update=False)

        update_object(wildcard_domain, lambda o, d: o.add_node(d), new_node)
        res = result_schema.dump(
            Result(
                message="Created new node {}".format(new_node["id"]),
                data=[node_schema.dump(new_node)],
            )
        )

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    g.created_objects.append(new_node["id"])

    return response


@blueprint.route(  # type: ignore
    "/nodes/<uuid:node_id>",
    methods=["GET"],
    summary="Get node",
    path_parameters={"node_id": "Id of the node"},
    operationId="getNode",
)
@result_data(
    description="Get node response",
    schemas=[node_schema],
    max_items=1,
)
@authenticated
@requires_endpoint_visibility
def nodes_get_by_id(node_id: uuid.UUID) -> Result:
    """
    Get information about one of the endpoint's node.

    This operation requires either the endpoint to have the public property set
    to true or the caller to have a role that allows viewing this Endpoint or a
    Collection on it.
    """
    current_app.logger.debug(f"Attempting to look up node {node_id}")

    wildcard_domain = m.GlobusDnsDomain.query.one_or_none()

    if not wildcard_domain:
        raise errors.InternalServerError("Server is misconfigured")

    data = []
    for node in wildcard_domain.nodes:
        if node.id == str(node_id):
            data.append(node_schema.dump(node))
            break

    if not data:
        raise errors.NotFoundError()

    res = result_schema.dump(Result(data=data))

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


class NodeUpdateInclude(enum.Enum):
    node = "node"


@blueprint.route(
    "/nodes/<uuid:node_id>",
    methods=["PUT"],
    summary="Update a node",
    path_parameters={"node_id": "Id of the node"},
    operationId="putNode",
)
@result_data(
    description="Update node response",
    schemas=[node_schema],
    max_items=1,
)
@authenticated
@query_parameter(
    name_spec="include",
    description="""
    Flag indicating whether to request all roles assignments for the endpoint
    or collection.
    """,
    converter=enum_list_converter(NodeUpdateInclude),
    default=[],
)
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
@expects_json(
    schema=node_schema,
    description="New node definition",
    required=True,
)
def put_node(*, node_id: uuid.UUID, include: t.List[NodeUpdateInclude]) -> Result:
    """
    Update a node, replacing all properties with those in the input.  This
    operation optionally returns the node's definition after the update if the
    `include=node` query parameter is passed to this operation.
    """
    node_id_str = str(node_id)
    include_node = NodeUpdateInclude.node in include

    current_app.logger.debug(
        "Attempting to update node {} with {}".format(node_id_str, g.validated_data)
    )

    if node_id_str != str(g.validated_data["id"]):
        raise errors.BadRequestError()

    with current_app.appsync_client.pause_notifications():
        wildcard_domain = m.GlobusDnsDomain.query.one_or_none()

        if not wildcard_domain:
            raise errors.InternalServerError("Server WILDCARD_DOMAIN is not configured")

        if node_id_str not in [n.id for n in wildcard_domain.nodes]:
            raise errors.NotFoundError()

        _check_for_node_conflicts(g.validated_data, wildcard_domain)

        update_object(wildcard_domain, lambda o, d: o.put_node(d), g.validated_data)

        data = []
        if include_node:
            data.append(node_schema.dump(wildcard_domain.get_node(id=node_id_str)))
        res = result_schema.dump(
            Result(message="Updated node {}".format(node_id_str), data=data)
        )

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    g.updated_objects.append(node_id_str)

    return response


@blueprint.route(
    "/nodes/<uuid:node_id>",
    methods=["PATCH"],
    summary="Update a node",
    path_parameters={"node_id": "Id of the node"},
    operationId="patchNode",
)
@result_data(
    description="Update node response",
    schemas=[node_schema],
    max_items=1,
)
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
@query_parameter(
    name_spec="include",
    description="""
    Flag indicating whether to request all roles assignments for the endpoint
    or collection.
    """,
    converter=enum_list_converter(NodeUpdateInclude),
    default=[],
)
@expects_json(
    schema=node_patch_schema,
    description="Updates to node definition",
    required=True,
)
def patch_node(node_id: uuid.UUID, include: t.List[NodeUpdateInclude]) -> Result:
    """
    Update a node, changing only the properties included in the input document.
    Items explicitly set to null in the input are removed from the Node
    document. This operation optionally returns the node defintion after
    applying the changes in the input if the `include=node` query parameter is
    passed to this operation.
    """
    node_id_str = str(node_id)
    include_node = NodeUpdateInclude.node in include

    current_app.logger.debug(
        "Attempting to update node {} with {}".format(node_id_str, g.validated_data)
    )

    if "id" not in g.validated_data:
        g.validated_data["id"] = node_id_str

    if node_id_str != str(g.validated_data["id"]):
        raise errors.BadRequestError()

    with current_app.appsync_client.pause_notifications():
        wildcard_domain = m.GlobusDnsDomain.query.one_or_none()
        if not wildcard_domain:
            raise errors.InternalServerError("Server is misconfigured")

        _check_for_node_conflicts(g.validated_data, wildcard_domain)

        update_object(wildcard_domain, lambda o, d: o.patch_node(d), g.validated_data)

        data = []
        if include_node:
            data.append(node_schema.dump(wildcard_domain.get_node(id=node_id_str)))

        res = result_schema.dump(
            Result(message="Updated node {}".format(node_id_str), data=data)
        )

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    g.updated_objects.append(node_id_str)

    return response


@blueprint.route(
    "/nodes/<uuid:node_id>",
    methods=["DELETE"],
    summary="Delete a node",
    path_parameters={"node_id": "Id of the node"},
    operationId="deleteNode",
)
@result_data(
    description="Delete node response",
    schemas=[],
)
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
def delete_node(node_id):
    """
    Delete the `Node` document for the given node.
    """
    node_id_str = str(node_id)

    current_app.logger.debug("Attempting to delete node %s", node_id_str)

    with current_app.appsync_client.pause_notifications():
        wildcard_domain = m.GlobusDnsDomain.query.one_or_none()

        if not wildcard_domain:
            raise errors.InternalServerError("Server WILDCARD_DOMAIN is not configured")

        if node_id_str not in [n.id for n in wildcard_domain.nodes]:
            raise errors.NotFoundError()

        update_object(
            wildcard_domain, lambda o, **kwargs: o.remove_node(**kwargs), id=node_id_str
        )

        res = result_schema.dump(Result(message="Deleted node {}".format(node_id_str)))

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    g.deleted_objects.append(node_id_str)

    return response


def _check_for_node_conflicts(node_document, wildcard_domain, update=True):
    """
    Verify that the node changes in the node document would not create any
    conflicts with other nodes in this wildcard domain.

    :param node_document:
        A dict containing a node document with new values for this domain. This
        may be a complete document (from a POST or PUT) or partial (from a
        PATCH).
    :param wildcard_domain:
        A GlobusDnsDomain model object containing the current node information for this
        endpoint.
    :param update:
        Boolean flag indicating whether this is checking at creation or
        updating. If True, require the node id in the node document to exist
        in the domain before this function is called.

    :raises errors.ResourceExistsError:
        If any node conflicts with this new node document, this function raises
        an exception containing information include the node UUIDs and IP
        addresses which conflict with the new node.
    """
    node_id_str = str(node_document["id"])

    original_node = None
    for node in wildcard_domain.nodes:
        if node.id == node_id_str:
            original_node = node
            break

    if original_node is None and update:
        raise errors.NotFoundError(
            "Node {} does not exist on this endpoint".format(node_document["id"])
        )

    ip_addresses = [str(ip) for ip in node_document.get("ip_addresses", [])]
    existing_ips = [str(ip) for ip in wildcard_domain.ip_addresses()]
    conflicting_nodes = set()
    conflicting_ips = set()
    for ip in ip_addresses:
        if ip in existing_ips:
            conflicting_node = wildcard_domain.get_node(ip_address=ip)
            if original_node is None or conflicting_node.id != original_node.id:
                conflicting_ips.add(ip)
                conflicting_nodes.add(conflicting_node.id)

    if conflicting_nodes:
        raise errors.ResourceExistsError(
            error_description=(
                "A node with address(es) {} already exists on "
                "this endpoint".format(", ".join(list(conflicting_ips)))
            ),
            resource_ids=list(conflicting_nodes),
        )
