from collections.abc import Mapping, Sequence
import datetime
import os
import re
from typing import Any, Dict, List, Optional, Set, Union
import uuid

from flask import current_app, g, render_template
import globus_sdk
from globus_sdk.scopes import AuthScopes, GCSCollectionScopeBuilder, TransferScopes

from sqlalchemy import and_, case, or_, select
from sqlalchemy.sql import func
from sqlalchemy.orm import aliased
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.ext.mutable import MutableDict

from globus.manager.api.utils import child_client
from globus.manager.api.utils.log import log_auth_exception, log_transfer_exception
from globus.manager.utils.run_as_user import run_as_user
from globus.manager.version import __version__

from . import (
    db,
    ConfigurationObject,
    Domain,
    LetsEncryptCertificate,
    Role,
    SharingPolicy,
    StorageGateway,
    UserCredential,
)
from globus.manager.utils.path_restriction import PathRestrictions
from .decorators import DocumentChecker, checks_property

from globus.manager.constants import CollectionType, ConfigurationObjectTag
from globus.manager.api.schemas.api import (
    DomainDict,
    SharingPolicyLoadDict,
    collection_configuration_schema,
    path_restrictions_schema,
    sharing_policy_schema,
)
from globus.manager.api.schemas.transfer import (
    gcs_owner_string_schema,
    mapped_collection_create_schema,
    mapped_collection_update_schema,
    guest_collection_create_schema,
    guest_collection_update_schema,
)
import globus.manager.api.errors as errors


class Collection(ConfigurationObject, DocumentChecker):
    """
    Collection model for mapped and guest collections in GCSv5
    """

    __tablename__ = "collection"

    @classmethod
    def polymorphic_class(cls, discriminator):
        return ConfigurationObject.polymorphic_class(
            ConfigurationObjectTag.COLLECTION.name, discriminator
        )

    COLLECTION_PROPERTIES = {}
    COLLECTION_PROPERTIES["collection#1.0.0"] = [
        # Columns
        "allow_guest_collections",
        "collection_type",
        "mapped_collection_id",
        "storage_gateway_id",
        "user_credential_id",
        # Configuration hash
        "collection_base_path",
        "contact_email",
        "contact_info",
        "default_directory",
        "deleted",
        "department",
        "description",
        "disable_verify",
        "display_name",
        "domain_name",
        "identity_id",
        "info_link",
        "keywords",
        "organization",
        "policies",
        "public",
        "sharing_restrict_paths",
        "force_encryption",
    ]
    COLLECTION_PROPERTIES["collection#1.1.0"] = COLLECTION_PROPERTIES[
        "collection#1.0.0"
    ] + ["enable_https", "user_message", "user_message_link"]
    COLLECTION_PROPERTIES["collection#1.2.0"] = COLLECTION_PROPERTIES[
        "collection#1.1.0"
    ] + ["sharing_users_allow", "sharing_users_deny"]
    COLLECTION_PROPERTIES["collection#1.3.0"] = COLLECTION_PROPERTIES[
        "collection#1.2.0"
    ]
    COLLECTION_PROPERTIES["collection#1.4.0"] = COLLECTION_PROPERTIES[
        "collection#1.3.0"
    ] + ["force_verify"]
    COLLECTION_PROPERTIES["collection#1.5.0"] = COLLECTION_PROPERTIES[
        "collection#1.4.0"
    ] + ["disable_anonymous_writes"]
    COLLECTION_PROPERTIES["collection#1.6.0"] = COLLECTION_PROPERTIES[
        "collection#1.5.0"
    ] + ["guest_auth_policy_id"]
    COLLECTION_PROPERTIES["collection#1.7.0"] = COLLECTION_PROPERTIES[
        "collection#1.6.0"
    ]
    COLLECTION_PROPERTIES["collection#1.8.0"] = COLLECTION_PROPERTIES[
        "collection#1.7.0"
    ] + ["delete_protected"]
    # 1.9 Adds last_access and created_at but those are read-only
    COLLECTION_PROPERTIES["collection#1.9.0"] = (
        COLLECTION_PROPERTIES["collection#1.8.0"] + []
    )

    _DATA_TYPE = "collection#1.9.0"
    _DATA_TYPES = list(COLLECTION_PROPERTIES.keys())

    CONNECTOR_POLICIES: Dict[str, List[str]] = {}
    """
    Connector-specific mapping of DATA_TYPE to list of properties which can
    be present in the policy document
    """

    configuration_schema = collection_configuration_schema

    # Column definitions
    id = db.Column(
        db.TEXT,
        db.ForeignKey("configuration_object.id", name="col_id"),
        primary_key=True,
    )
    _domain_name = db.Column("domain", db.TEXT, nullable=False, unique=True)
    _user_credential_id = db.Column(
        "user_credential_id",
        db.TEXT,
        db.ForeignKey("user_credential.id"),
        nullable=True,
    )
    user_credential = db.relationship(
        "UserCredential",
        foreign_keys=[_user_credential_id],
        uselist=False,
        lazy="joined",
    )
    _storage_gateway_id = db.Column(
        "storage_gateway_id",
        db.TEXT,
        db.ForeignKey("storage_gateway.id"),
        nullable=False,
    )
    _collection_type = db.Column(
        "collection_type", db.Enum(CollectionType), nullable=False
    )
    _allow_guest_collections = db.Column(
        "allow_guest_collections", db.BOOLEAN, nullable=False, server_default="0"
    )
    _mapped_collection_id = db.Column(
        "mapped_collection_id", db.TEXT, db.ForeignKey("collection.id"), nullable=True
    )
    storage_gateway = db.relationship(
        "StorageGateway",
        foreign_keys=[_storage_gateway_id],
        uselist=False,
        lazy="joined",
    )
    mapped_collection = db.relationship(
        "ConfigurationObject",
        primaryjoin="remote(ConfigurationObject.id) "
        "== foreign(Collection.mapped_collection_id)",
        uselist=False,
        viewonly=True,
        lazy="joined",
    )
    owner_role = db.relationship(
        "Role",
        primaryjoin="and_(Collection.id==Role.resource, Role.primary_role)",
        uselist=False,
    )
    roles = db.relationship(
        "Role",
        primaryjoin="Collection.id == Role.resource",
        uselist=True,
        backref="collection",
    )
    endpoint = db.relationship(
        "Endpoint",
        primaryjoin="remote(Endpoint.id) != foreign(Collection.id)",
        uselist=False,
        viewonly=True,
    )
    sharing_policies = db.relationship(
        "SharingPolicy",
        primaryjoin="Collection.id==SharingPolicy._collection_id",
        back_populates="collection",
        cascade="all, save-update, delete, delete-orphan",
        lazy="joined",
    )

    _domain: Optional[Domain] = None
    last_access = db.Column("last_access", db.Date, nullable=True)

    @hybrid_property
    def created_at_date(self) -> datetime.date:
        return self.created_at.date()

    @created_at_date.expression  # type: ignore
    def created_at_date(cls):
        # This is calling the SQLite DATE function on the column, not portable
        # to other DBs
        return func.date(cls.created_at)

    # methods and properties
    def __init__(
        self,
        appsync_data=None,
        doc=None,
        storage_gateway=None,
        mapped_collection=None,
        user_credential=None,
        endpoint=None,
    ):

        if appsync_data is not None:
            args = {k: v for k, v in appsync_data.items()}

            args["tag"] = ConfigurationObjectTag.COLLECTION
            args["secondary_discriminator"] = self.connector_id

            # If the collection is not hard-deleted, we should rely on the deleted status
            # within the configuration hash. If we use both, there's a random chance that
            # db.Model will initialize the configuration object in an order (configuration
            # before item['deleted']) which causes the object to be marked as not deleted.
            if not args.get("deleted", True):
                args.pop("deleted")

            if mapped_collection:
                self.mapped_collection = mapped_collection
            if storage_gateway:
                self.storage_gateway = storage_gateway
            if user_credential:
                self.user_credential = user_credential
            if endpoint:
                self.endpoint = endpoint

            super().__init__(**args)
            self.init_appsync()
        elif doc.get("DATA_TYPE") in self._DATA_TYPES:
            if doc.get("enable_https") and not current_app.subscription:
                raise errors.SubscriptionRequiredError(
                    "The enable_https property may not be set on an unmanaged endpoint",
                    subscription_level="starter",
                )
            self.configuration_hash = MutableDict()
            self.collection_type = doc.get("collection_type")
            self.user_credential_id = doc.get("user_credential_id")
            self.mapped_collection_id = doc.get("mapped_collection_id")
            self.storage_gateway_id = doc.get("storage_gateway_id")

            self.display_name = doc.get("display_name")

            if not doc.get("storage_gateway_id"):
                raise ValueError("Missing required storage_gateway_id")

            if self.is_guest_collection and not doc.get("mapped_collection_id"):
                raise ValueError(
                    "Missing required mapped_collection_id for guest collection"
                )

            # Seed relationships on initialization
            if self.configuration_hash.get("mapped_collection_id", None):
                if mapped_collection:
                    assert mapped_collection.id == self.mapped_collection_id
                    self.mapped_collection = mapped_collection

            if storage_gateway:
                assert storage_gateway.id == self._storage_gateway_id
                self.storage_gateway = storage_gateway

            if endpoint:
                self.endpoint = endpoint

            if not doc.get("identity_id"):
                raise ValueError("Missing required identity_id")

            if self._user_credential_id:
                if not user_credential:
                    raise ValueError("Invalid user_credential_id")
                if self.storage_gateway_id is None:
                    self.storage_gateway_id = user_credential.storage_gateway_id
                elif user_credential.storage_gateway_id != self.storage_gateway_id:
                    raise ValueError("Invalid storage_gateway_id")
            elif self.is_guest_collection and self.storage_gateway.require_credential:
                raise ValueError("Missing required user_credential")

            self.domain_name = doc["domain_name"]
            self.id = self.register_child_client()
            id_str = self.id
            super().__init__(
                id=id_str,
                tag=ConfigurationObjectTag.COLLECTION,
                secondary_discriminator=self.connector_id,
            )
            self.post(**doc)
            self.refresh_configuration_hash()
            try:
                admin_role = None
                transfer_doc = None

                transfer_doc = self.create_transfer_endpoint()
                if transfer_doc["id"] != self.id:
                    raise errors.InternalServerError("Bad response from transfer")

                if self.is_guest_collection or current_app.subscription:
                    admin_role = Role(
                        doc=dict(
                            DATA_TYPE="role#1.0.0",
                            collection=self.id,
                            principal="urn:globus:auth:identity:{}".format(
                                self.identity_id
                            ),
                            role="administrator",
                        ),
                        primary_role=True,
                    )
                self.owner_role = admin_role
            except Exception as e:
                current_app.logger.warning("Error creating resource: %s", e)
                if admin_role:
                    admin_role.delete()
                if transfer_doc:
                    self.delete_transfer_endpoint(endpoint_id=transfer_doc["id"])
                self.unregister_child_client()
                raise
        else:
            raise ValueError("Unsupported DATA_TYPE")

    def init_appsync(self) -> None:
        if self.configuration_hash:
            self._domain = None
            if self.configuration_hash.get("policies") is None:
                self.configuration_hash["policies"] = {
                    "DATA_TYPE": self.POLICY_DATA_TYPE
                }

            if self.configuration_hash.get("last_access") is not None:
                try:
                    self.last_access = datetime.date.fromisoformat(
                        self.configuration_hash["last_access"]
                    )
                except ValueError:
                    current_app.logger.error(
                        "Ignoring invalid last_access value in appsync data"
                    )

            self.id = self.configuration_hash["id"]
            # mypy can't handle @hybrid_property assignments, hence the type ignores
            self.collection_type = self.configuration_hash["collection_type"]  # type: ignore

            # [SC-23049, 5.4.59]: On DELETE, 'domain_name' may be missing from the
            # hash but the 'domain' collection table field is accurate. So if
            # 'domain_name' is missing from hash, don't set the 'domain_name'
            # attribute; allow it to fallback on the DB 'domain' column.
            if "domain_name" in self.configuration_hash:
                self.domain_name = self.configuration_hash["domain_name"]  # type: ignore

            self.user_credential_id = self.configuration_hash.get("user_credential_id")  # type: ignore
            self.storage_gateway_id = self.configuration_hash["storage_gateway_id"]  # type: ignore
            self.allow_guest_collections = self.configuration_hash.get(  # type: ignore
                "allow_guest_collections"
            )
            self.mapped_collection_id = self.configuration_hash.get(  # type: ignore
                "mapped_collection_id"
            )
            if self.configuration_hash.get("deleted"):
                self.deleted = True  # type: ignore
            if self.is_mapped_collection and not self.configuration_hash.get(
                "root_path"
            ):
                self.configuration_hash["root_path"] = self.collection_base_path

            if self.is_mapped_collection:
                new_policy_dict: Optional[SharingPolicyLoadDict] = None
                appsync_policies: Dict[str, SharingPolicyLoadDict] = {}

                # Only reconcile sharing_policies for non-deleted collections. For deleted
                # collections, we'll treat it as if appsync_policies is empty and every
                # sharing policy needs to be deleted.
                if not self.deleted:
                    if self.configuration_hash.get(
                        "sharing_restrict_paths"
                    ) and not self.configuration_hash.get("sharing_policies"):
                        # If sharing_policies is missing but sharing_restrict_paths
                        # exists, update the configuration_hash to have the former.
                        policy_id = self._default_sharing_policy_id
                        new_policy_dict = {
                            "DATA_TYPE": "sharing_policy#1.0.0",
                            "id": uuid.UUID(policy_id),
                            "collection_id": uuid.UUID(self.id),
                            "users": None,
                            "sharing_restrict_paths": self.configuration_hash[
                                "sharing_restrict_paths"
                            ],
                        }
                        del self.configuration_hash["sharing_restrict_paths"]
                    elif not self.configuration_hash.get("sharing_policies"):
                        # If neither sharing_policies nor sharing_restrict_paths exist
                        # in the configuration_hash, create a default policy that
                        # is RW/ for any user. This is the default value, but only
                        # comes into play if allow_guest_collections is True
                        policy_id = self._default_sharing_policy_id
                        new_policy_dict = {
                            "DATA_TYPE": "sharing_policy#1.0.0",
                            "id": uuid.UUID(policy_id),
                            "collection_id": uuid.UUID(self.id),
                            "users": None,
                            "sharing_restrict_paths": path_restrictions_schema.dump(
                                PathRestrictions(read_write=["/"])
                            ),
                        }

                    appsync_policies = {
                        str(sp["id"]): sp
                        for sp in self.configuration_hash.get("sharing_policies", [])
                    }

                    if new_policy_dict:
                        appsync_policies[policy_id] = new_policy_dict

                model_policies: Dict[str, SharingPolicy] = {
                    sp.id: sp for sp in self.sharing_policies
                }

                for spid, doc in appsync_policies.items():
                    if spid not in model_policies:
                        new_policy = SharingPolicy(doc=doc, collection=self)
                        model_policies[new_policy.id] = new_policy
                    else:
                        model_policies[spid].post(doc=doc)
                for spid, policy in model_policies.items():
                    if spid not in appsync_policies:
                        self.sharing_policies.remove(policy)
                self.refresh_configuration_hash()

    @property
    def manager_url(self):
        """
        HTTPS URL of the GCS Manager service which manages this collection.
        """
        if self.is_deleted:
            raise AttributeError("This collection is deleted")
        return "https://{}".format(self.endpoint.domain_name)

    @checks_property(prop="manager_url")
    def check_manager_url(self, value: str):
        old_value = self.configuration_hash.get("manager_url", None)
        if old_value and value != old_value:
            raise AttributeError()

    @property
    def https_url(self):
        """
        URL using the "https" scheme which can be used by web clients to
        access data on this collection.
        """
        if self.is_deleted:
            raise AttributeError("This collection is deleted")

        if not current_app.subscription or not self.enable_https:
            return None

        return "https://{}".format(self.domain_name)

    @checks_property(prop="https_url")
    def check_https_url(self, value):
        old_value = self.configuration_hash.get("https_url", None)
        if old_value and value != old_value:
            raise AttributeError()

    @property
    def tlsftp_url(self):
        """
        URL using the "tslftp" scheme which is used by the Globus Transfer
        service to access data on this collection.
        """
        if self.is_deleted:
            raise AttributeError("This collection is deleted")
        return (
            f"tlsftp://{self.domain_name}:{self.endpoint.gridftp_control_channel_port}"
        )

    @checks_property(prop="tlsftp_url")
    def check_tlsftp_url(self, value):
        old_value = self.configuration_hash.get("tlsftp_url", None)
        if old_value and value != old_value:
            raise AttributeError()

    @hybrid_property  # type: ignore
    def collection_type(self) -> str:
        """
        Type of this collection: currently one of "mapped" or "guest". This may
        not change after the collection is created.
        """
        if self._collection_type is not None:
            return CollectionType(self._collection_type).name

    @collection_type.expression  # type: ignore
    def collection_type(self):
        return self._collection_type

    @collection_type.setter  # type: ignore
    def collection_type(self, value):
        new_value = CollectionType(value).name
        if new_value != self._collection_type:
            self._collection_type = new_value
            if new_value != self.configuration_hash.get("collection_type"):
                self.configuration_hash["collection_type"] = new_value

    @checks_property(prop="collection_type")
    def check_collection_type(self, value: Union[str, CollectionType]) -> None:
        old_collection_type = self.configuration_hash.get("collection_type", None)
        if (
            old_collection_type
            and CollectionType(old_collection_type).name != CollectionType(value).name
        ):
            raise AttributeError()

    @checks_property(prop="connector_id")
    def check_connector_id(self, value: str):
        if value and str(value) != self.connector_id:
            raise AttributeError()

    @property
    def display_name(self):
        """
        The user-friendly name for this collection. This may be a Unicode
        string of at most 128 characters with no newline characters.
        """
        return self.configuration_hash["display_name"]

    @display_name.setter
    def display_name(self, value: str):
        if value and value != self.configuration_hash.get("display_name"):
            self.configuration_hash["display_name"] = value

    @hybrid_property
    def storage_gateway_id(self):
        return self._storage_gateway_id

    @storage_gateway_id.setter  # type: ignore
    def storage_gateway_id(self, value):
        if value:
            if self._storage_gateway_id != value:
                self._storage_gateway_id = str(value)
            if self.configuration_hash.get("storage_gateway_id") != value:
                self.configuration_hash["storage_gateway_id"] = str(value)

    @checks_property(prop="storage_gateway_id", required=True)
    def check_storage_gateway_id(self, value: str) -> None:
        if self._storage_gateway_id is not None and self._storage_gateway_id != value:
            raise AttributeError()

    @hybrid_property
    def user_credential_id(self):
        if self._user_credential_id:
            return self._user_credential_id
        else:
            raise AttributeError(
                "The user_credential_id property is not defined for this collection"
            )

    @user_credential_id.setter  # type: ignore
    def user_credential_id(self, value: Optional[str]):
        self._user_credential_id = value

    @checks_property(prop="user_credential_id")
    def check_user_credential_id(self, value: str) -> None:
        if self.is_mapped_collection:
            if value:
                raise AttributeError(
                    "The user_credential_id property may not be set on mapped collections"
                )
            else:
                return

        old_ucid = self.configuration_hash.get("user_credential_id", None)

        if old_ucid is not None and old_ucid != value:
            raise AttributeError()

    def update_last_access(self, value: Optional[datetime.date]) -> None:
        if value:
            old_value: Optional[datetime.date] = self.last_access
            if old_value is None or value > old_value:
                try:
                    self.last_access = value
                    self.update_transfer_endpoint()
                except Exception:
                    self.last_access = old_value

    def update_domain_name(self, new_domain: str, force: bool = False) -> None:
        """
        Rewrite the domain name, because a custom domain change occurred. If
        this changes things, also update the transfer endpoint, child client,
        and scopes
        """
        before = self._domain_name
        new_label = new_domain.split(".", 1)[0]

        if self.is_guest_collection:
            assert self.mapped_collection is not None

        if self._domain is not None:
            new_name = self._domain.domain_name
        elif self.is_guest_collection and self.mapped_collection.has_wildcard_domain:
            new_name = ".".join([new_label, self.mapped_collection.domain.domain_name])
        elif self.endpoint.has_wildcard_domain:
            new_name = ".".join([new_label, self.endpoint.domain.domain_name])
        else:
            new_name = ".".join(
                [new_label, self.endpoint.globus_dns_domain.domain_name]
            )
        self._domain_name = new_name

        if self._domain_name != before or force:
            self.refresh_configuration_hash()
            self.update_transfer_endpoint()
            if not self.is_deleted:
                self.update_child_client()
                self.update_scopes()

    @hybrid_property
    def domain_name(self) -> str:
        """
        Fully-qualified domain name of the collection. For a mapped collection,
        This may be updated, but is immutable for guest collections.
        """
        if self.is_deleted:
            return self.id + "-[DELETED]"

        label, domain = self._domain_name.split(".", 1)

        if self.has_domain:
            return self.domain.domain_name
        elif self.is_guest_collection and self.mapped_collection.has_wildcard_domain:
            return ".".join([label, self.mapped_collection.domain.domain_name])
        elif self.endpoint.has_wildcard_domain:
            return ".".join([label, self.endpoint.domain.domain_name])
        else:
            return ".".join([label, self.endpoint.globus_dns_domain.domain_name])

    @domain_name.expression  # type: ignore
    def domain_name(cls) -> str:
        return cls._domain_name

    @domain_name.setter  # type: ignore
    def domain_name(self, value):
        old_name = self._domain_name
        if old_name and old_name == value:
            return
        self._domain_name = value
        self.configuration_hash["domain_name"] = value

    @checks_property(prop="domain_name")
    def check_domain_name(self, value: str) -> None:
        if value is None:
            raise ValueError("Invalid value for domain_name property.")

        domain_name = self.configuration_hash.get("domain_name", None)
        if (
            domain_name is not None
            and self.is_guest_collection
            and value != self.domain_name
        ):
            raise ValueError(
                "The domain_name property may not be changed for guest collections."
            )

    @property
    def identity_id(self):
        return self.configuration_hash["identity_id"]

    @identity_id.setter
    def identity_id(self, value):
        self.configuration_hash["identity_id"] = str(value)

    @checks_property(prop="identity_id")
    def check_identity_id(self, value: str) -> None:
        old_identity_id = self.configuration_hash.get("identity_id", None)

        if old_identity_id is not None:
            if value != old_identity_id:
                raise AttributeError()
        else:
            try:
                if self.is_guest_collection:
                    uc = self.user_credential
                    if uc:
                        if uc.identity_id != value:
                            raise ValueError(
                                "The identity_id property must match that of the "
                                "user_credential associated with this collection."
                            )
            except AttributeError:
                pass

    @checks_property(prop="delete_protected")
    def check_delete_protected(self, value: Optional[bool]) -> None:
        if self.is_guest_collection and value is not None:
            raise AttributeError(
                "Delete protection not supported for guest collections"
            )

    @property
    def root_path(self):
        if self.is_guest_collection:
            return re.sub(
                "//*",
                "/",
                "/".join([self.mapped_collection.root_path, self.collection_base_path]),
            )
        else:
            return self.configuration_hash["root_path"]

    @property
    def collection_base_path(self):
        value = self.configuration_hash.get("collection_base_path")
        if value is not None:
            return value
        raise AttributeError(
            "The collection_base_path property is not set on this collection"
        )

    @collection_base_path.setter
    def collection_base_path(self, value):
        self.configuration_hash["collection_base_path"] = value

        if self.is_mapped_collection:
            self.configuration_hash["root_path"] = value

    @checks_property(prop="collection_base_path")
    def check_collection_base_path(self, value: str) -> None:
        old_value = self.configuration_hash.get("collection_base_path", None)
        if old_value is not None and value != old_value:
            raise AttributeError()

        if (
            self.is_guest_collection
            and self.mapped_collection.collection_base_path != "/"
        ):
            for bad in ["$HOME", "~"]:
                if bad in value:
                    raise ValueError(
                        "{} not allowed in guest collection base path.".format(bad)
                    )

    @property
    def default_directory(self):
        value = self.configuration_hash.get("default_directory")
        if value:
            return value
        raise AttributeError(
            "The default_directory property is not set for this collection"
        )

    @default_directory.setter
    def default_directory(self, value):
        self.configuration_hash["default_directory"] = value

    @default_directory.deleter
    def default_directory(self):
        self.configuration_hash["default_directory"] = "/"

    @checks_property(prop="default_directory")
    def check_default_directory(self, value: str) -> None:
        """
        Only allow $HOME and ~ on mapped collections when base_path is "/"; we never
        allow it on guest collections. If base_path is not set yet, then we are being
        created and post() will handle the check for us.
        """
        base_path = self.configuration_hash.get("collection_base_path", None)
        if not self.is_mapped_collection or base_path not in [None, "/"]:
            for bad in ["$HOME", "~"]:
                if value and bad in value:
                    raise ValueError(
                        "{} not allowed in collection default_directory.".format(bad)
                    )

    public = ConfigurationObject.model_property("public", required=True)

    @property
    def high_assurance(self):
        return self.storage_gateway.high_assurance

    @property
    def force_encryption(self):
        if bool(self.storage_gateway.high_assurance):
            return True
        if self.is_guest_collection:
            if self.mapped_collection.force_encryption:
                return True
        return self.configuration_hash.get("force_encryption", False)

    @force_encryption.setter
    def force_encryption(self, value):
        if self.is_mapped_collection and not self.storage_gateway.high_assurance:
            self.configuration_hash["force_encryption"] = bool(value)
        elif self.is_guest_collection and not self.mapped_collection.force_encryption:
            self.configuration_hash["force_encryption"] = bool(value)

    @force_encryption.deleter
    def force_encryption(self):
        if self.storage_gateway.high_assurance:
            self.configuration_hash["force_encryption"] = True
        elif self.is_mapped_collection:
            self.configuration_hash["force_encryption"] = False
        else:
            self.configuration_hash[
                "force_encryption"
            ] = self.mapped_collection.force_encryption

    @checks_property(prop="force_encryption")
    def check_force_encryption(self, value: bool) -> None:
        # If the caller explicitly disables force_encryption...
        if value is False:
            if self.storage_gateway.high_assurance:
                raise ValueError(
                    "The force_encryption property must be true for high assurance "
                    "collections."
                )
            if self.is_guest_collection and self.mapped_collection.force_encryption:
                raise ValueError(
                    "The force_encryption property must be true for guest "
                    "collections when the mapped collection requires "
                    "encryption."
                )

    @property
    def disable_anonymous_writes(self):
        # HA collections do not allow anonymous access (ever)
        if self.storage_gateway.high_assurance:
            return True
        if self.is_guest_collection:
            return self.mapped_collection.disable_anonymous_writes
        return self.configuration_hash.get("disable_anonymous_writes", False)

    @disable_anonymous_writes.setter
    def disable_anonymous_writes(self, value):
        if self.is_mapped_collection:
            self.configuration_hash["disable_anonymous_writes"] = bool(value)

    @disable_anonymous_writes.deleter
    def disable_anonymous_writes(self):
        if self.is_mapped_collection:
            self.configuration_hash.pop("disable_anonymous_writes", None)

    @property
    def delete_protected(self):
        if self.is_guest_collection:
            raise AttributeError("delete_protected")
        return self.configuration_hash.get("delete_protected", False)

    @delete_protected.setter
    def delete_protected(self, value):
        if self.is_mapped_collection:
            self.configuration_hash["delete_protected"] = bool(value)

    @delete_protected.deleter
    def delete_protected(self):
        if self.is_mapped_collection:
            self.configuration_hash.pop("delete_protected", None)

    @checks_property(prop="disable_anonymous_writes")
    def check_disable_anonymous_writes(self, value):
        # Only display the error when the value is being changed. This allows
        # the test code to schema dump the model to test the api.
        if bool(value) != self.disable_anonymous_writes:
            # High assurance gateways have no anonymous access (ever)
            if self.storage_gateway.high_assurance and not bool(value):
                raise ValueError(
                    "The disable_anonymous_writes property can not be enabled "
                    "on high assurance collections."
                )

            # This flag is read only on guest collections
            if self.is_guest_collection:
                raise ValueError(
                    "The disable_anonymous_writes property can only be set "
                    "on non high assurance mapped collections."
                )

    @property
    def disable_verify(self):
        if self.is_guest_collection:
            if self.mapped_collection.disable_verify:
                return True
        return self.configuration_hash.get("disable_verify", False)

    @disable_verify.setter
    def disable_verify(self, value):
        if self.is_mapped_collection or not self.mapped_collection.disable_verify:
            self.configuration_hash["disable_verify"] = bool(value)

    @disable_verify.deleter
    def disable_verify(self):
        if self.is_mapped_collection or not self.mapped_collection.disable_verify:
            self.configuration_hash["disable_verify"] = False

    @checks_property(prop="disable_verify")
    def check_disable_verify(self, value: bool) -> None:
        if value is False:
            if self.is_guest_collection and self.mapped_collection.disable_verify:
                raise ValueError(
                    "The disable_verify property must be true for guest "
                    "collections when the mapped collection disables "
                    "verify."
                )

    @property
    def force_verify(self):
        if self.is_guest_collection:
            if self.mapped_collection.force_verify:
                return True
        return self.configuration_hash.get("force_verify", False)

    @force_verify.setter
    def force_verify(self, value):
        if self.is_mapped_collection or not self.mapped_collection.force_verify:
            self.configuration_hash["force_verify"] = bool(value)

    @force_verify.deleter
    def force_verify(self):
        if self.is_mapped_collection or not self.mapped_collection.force_verify:
            self.configuration_hash["force_verify"] = False

    @checks_property(prop="force_verify")
    def check_force_verify(self, value: bool) -> None:
        if value is False:
            if self.is_guest_collection and self.mapped_collection.force_verify:
                raise ValueError(
                    "The force_verify property must be true for guest "
                    "collections when the mapped collection forces "
                    "verify."
                )

    @property
    def guest_auth_policy_id(self):
        if self.is_guest_collection:
            return self.mapped_collection.guest_auth_policy_id
        auth_policy_id = self.configuration_hash.get("guest_auth_policy_id")
        if auth_policy_id is not None:
            return auth_policy_id
        raise AttributeError(
            "The guest_auth_policy_id property is not defined for this collection."
        )

    @guest_auth_policy_id.setter
    def guest_auth_policy_id(self, value):
        if self.is_mapped_collection:
            self.configuration_hash["guest_auth_policy_id"] = str(value)

    @guest_auth_policy_id.deleter
    def guest_auth_policy_id(self):
        if self.is_mapped_collection:
            self.configuration_hash.pop("guest_auth_policy_id", None)

    @checks_property(prop="guest_auth_policy_id")
    def check_guest_auth_policy_id(self, value):
        # Only display the error when the value is being changed. This allows
        # the test code to schema dump the model to test the api.
        if value != getattr(self, "guest_auth_policy_id", None):
            # This flag is read only on guest collections
            if self.is_guest_collection:
                raise ValueError(
                    "The guest_auth_policy_id property can only be set on mapped collections."
                )
            if value is not None:
                # Only HA policies on HA collections and non-HA policies on non-HA collections
                try:
                    policy_data = current_app.auth_client.get_policy(value)
                except globus_sdk.AuthAPIError as e:
                    log_auth_exception(e)
                    if getattr(e, "code", None) == "NOT_FOUND":
                        raise ValueError(f"No policy exists with the ID {value}")
                    raise errors.BadRequestError(
                        getattr(e, "message", "Unable to assign policy to collection")
                    )

                if (
                    policy_data.data["high_assurance"]
                    != self.storage_gateway.high_assurance
                ):
                    raise ValueError(
                        "The high assurance setting for the policy must match the high assurance "
                        "setting on the collection's storage gateway."
                    )

    @hybrid_property
    def mapped_collection_id(self):
        if self.is_mapped_collection:
            raise AttributeError(
                "The mapped_collection_id property is not defined for this collection"
            )
        else:
            return self._mapped_collection_id

    @mapped_collection_id.expression  # type: ignore
    def mapped_collection_id(cls):
        return cls._mapped_collection_id

    @mapped_collection_id.setter  # type: ignore
    def mapped_collection_id(self, value):
        if self.is_mapped_collection:
            return
        old_value = self._mapped_collection_id
        if old_value is None and value:
            self._mapped_collection_id = str(value)
            self.configuration_hash["mapped_collection_id"] = str(value)

    @mapped_collection_id.deleter  # type: ignore
    def mapped_collection_id(self):
        if self.is_guest_collection:
            raise AttributeError(
                "The mapped_collection_id property is not defined for this collection"
            )

    @checks_property(prop="mapped_collection_id")
    def check_mapped_collection_id(self, value: Union[uuid.UUID, str]) -> None:
        if self.is_mapped_collection and value and str(value) != self.id:
            raise AttributeError(
                "The mapped_collection_id property may not be set on a mapped collection."
            )

        old_value = self.configuration_hash.get("mapped_collection_id", None)
        if old_value is not None and old_value != str(value):
            raise AttributeError("The mapped_collection_id property is immutable.")

    @hybrid_property
    def allow_guest_collections(self):
        if self.is_guest_collection:
            raise AttributeError(
                "The allow_guest_collections property is not defined for this collection"
            )

        if not current_app.subscription:
            return False

        return self._allow_guest_collections

    @allow_guest_collections.expression  # type: ignore
    def allow_guest_collections(cls):
        return cls._allow_guest_collections

    @allow_guest_collections.setter  # type: ignore
    def allow_guest_collections(self, value: bool):
        if not self.is_guest_collection:
            self._allow_guest_collections = value

    @allow_guest_collections.deleter  # type: ignore
    def allow_guest_collections(self):
        if self.is_mapped_collection:
            self._allow_guest_collections = False

    @checks_property(prop="allow_guest_collections")
    def check_allow_guest_collections(self, value: bool) -> None:
        if self.is_guest_collection:
            if value:
                raise AttributeError(
                    "The allow_guest_collections property can not be set on a guest collection."
                )
        elif not isinstance(value, bool):
            raise ValueError("Invalid value for the allow_guest_collections property.")

    @property
    def sharing_restrict_paths(self) -> PathRestrictions:
        if self.is_guest_collection:
            if self.user_credential:
                username = self.user_credential.username  # type: Any
            else:
                try:
                    identity_data = current_app.auth_client.get_identities(
                        ids=[self.identity_id]
                    )["identities"][0]
                    username = self.storage_gateway.map_identity(identity_data)
                except Exception:
                    raise AttributeError("Unable to resolve username")
            return self.absolute_sharing_restrict_paths(username)
        else:
            for policy in self.sharing_policies:
                if policy.is_default_policy:
                    return policy.sharing_restrict_paths

        raise AttributeError(
            "The sharing_restrict_paths property is not defined for this collection"
        )

    @sharing_restrict_paths.setter
    def sharing_restrict_paths(self, value):
        if self.is_mapped_collection:
            policy_id = self._default_sharing_policy_id
            restrictions = None

            if isinstance(value, Mapping):
                restrictions = PathRestrictions(**value)
            elif isinstance(value, PathRestrictions):
                restrictions = value
            else:
                restrictions = PathRestrictions(read_write=["/"])

            policy = None
            for p in self.sharing_policies:
                if p.id == policy_id:
                    policy = p
                    break

            if not policy:
                sharing_policy = SharingPolicy(
                    doc={
                        "DATA_TYPE": "sharing_policy#1.0.0",
                        "collection_id": self.id,
                        "sharing_restrict_paths": path_restrictions_schema.dump(
                            restrictions
                        ),
                        "users": None,
                    },
                    default_policy=True,
                )
                if sharing_policy not in self.sharing_policies:
                    self.sharing_policies.append(sharing_policy)
            else:
                policy.sharing_restrict_paths = restrictions

            self.configuration_hash["sharing_policies"] = sharing_policy_schema.dump(
                self.sharing_policies, many=True
            )

    @sharing_restrict_paths.deleter
    def sharing_restrict_paths(self):
        if self.is_mapped_collection:
            policy_id = self._default_sharing_policy_id
            policy = None
            for p in self.sharing_policies:
                if p.id == policy_id:
                    policy = p
                    break
            if policy:
                policy.sharing_restrict_paths = PathRestrictions(read_write=["/"])
            else:
                sharing_policy = SharingPolicy(
                    doc={
                        "DATA_TYPE": "sharing_policy#1.0.0",
                        "collection_id": self.id,
                        "sharing_restrict_paths": PathRestrictions(read_write=["/"]),
                        "users": None,
                    },
                    default_policy=True,
                )
                if sharing_policy not in self.sharing_policies:
                    self.sharing_policies.append(sharing_policy)

            self.configuration_hash["sharing_policies"] = sharing_policy_schema.dump(
                self.sharing_policies, many=True
            )

    @checks_property(prop="sharing_restrict_paths")
    def check_sharing_restrict_paths(
        self, value: Union[Mapping, PathRestrictions]
    ) -> None:
        if self.is_guest_collection:
            raise AttributeError(
                "The sharing_restrict_paths property may not be set on guest collections."
            )

        if not value:
            return

        if isinstance(value, Mapping):
            sharing_restrict_paths = PathRestrictions(**value)
        elif isinstance(value, PathRestrictions):
            sharing_restrict_paths = PathRestrictions(
                DATA_TYPE=value._DATA_TYPE,
                read=value.read,
                read_write=value.read_write,
                none=value.none,
            )
        else:
            raise ValueError()

        all_path_restrictions = (
            sharing_restrict_paths.read
            + sharing_restrict_paths.read_write
            + sharing_restrict_paths.none
        )

        for pr in all_path_restrictions:
            for bad in ["$HOME", "~"]:
                if bad in pr:
                    raise ValueError(
                        "{} not allowed in sharing_restrict_paths".format(bad)
                    )

    @property
    def sharing_users_allow(self) -> List[str]:
        """
        List of connector-specific usernames allowed to create new
        guest collections on this mapped collection. (**Added in
        API 1.2.0**)
        """
        value = self.configuration_hash.get("sharing_users_allow", None)
        if value:
            return value
        else:
            raise AttributeError("sharing_users_allow")

    @sharing_users_allow.setter
    def sharing_users_allow(self, value: Optional[List[str]]):
        if value != self.configuration_hash.get("sharing_users_allow"):
            self.configuration_hash["sharing_users_allow"] = value

    @sharing_users_allow.deleter
    def sharing_users_allow(self):
        if "sharing_users_allow" in self.configuration_hash:
            del self.configuration_hash["sharing_users_allow"]

    @checks_property(prop="sharing_users_allow")
    def _check_sharing_users_allow_value(self, value: Optional[List[str]]):
        if self.is_guest_collection:
            raise AttributeError()
        self.__validate_string_list("sharing_users_allow", value)

    @property
    def sharing_users_deny(self) -> Optional[List[str]]:
        """
        List of connector-specific usernames denyed to create new
        guest collections on this mapped collection. (**Added in
        API 1.2.0**)
        """
        value = self.configuration_hash.get("sharing_users_deny", None)
        if value:
            return value
        else:
            raise AttributeError("sharing_users_deny")

    @sharing_users_deny.setter
    def sharing_users_deny(self, value: Optional[List[str]]):
        if value != self.configuration_hash.get("sharing_users_deny"):
            self.configuration_hash["sharing_users_deny"] = value

    @sharing_users_deny.deleter
    def sharing_users_deny(self):
        if "sharing_users_deny" in self.configuration_hash:
            del self.configuration_hash["sharing_users_deny"]

    @checks_property(prop="sharing_users_deny")
    def _check_sharing_users_deny_value(self, value: Optional[List[str]]):
        if self.is_guest_collection:
            raise AttributeError()
        self.__validate_string_list("sharing_users_deny", value)

    @property
    def enable_https(self) -> bool:
        """
        Boolean flag indicating whether this collection should support HTTPS or
        not. This value can be set on mapped collections or guest collections.
        However, it may not be set to True if the value on the related mapped
        collection is False. (**Added in API 1.1.0**)
        """
        if self.is_guest_collection:
            assert self.mapped_collection is not None
            if self.mapped_collection.enable_https:
                return self.configuration_hash.get("enable_https", True)
            else:
                return False
        return self.configuration_hash.get("enable_https", True)

    @enable_https.setter
    def enable_https(self, value: bool):
        if self.is_mapped_collection or self.mapped_collection.enable_https:
            if bool(value) != self.configuration_hash.get("enable_https"):
                self.configuration_hash["enable_https"] = bool(value)

    @enable_https.deleter
    def enable_https(self, value: bool):
        self.configuration_hash["enable_https"] = True

    @checks_property(prop="enable_https")
    def check_enable_https(self, value):
        if not current_app.subscription and value:
            raise AttributeError()
        if self.is_guest_collection and value:
            if not self.mapped_collection.enable_https:
                raise AttributeError()

    _user_message = ConfigurationObject.model_property("user_message")

    @property
    def user_message(self) -> Optional[str]:
        """
        A message for clients to display to users when interacting with this
        endpoint.
        """
        if self.is_guest_collection:
            assert self.mapped_collection is not None
            return self.mapped_collection.user_message

        return self._user_message

    @user_message.setter
    def user_message(self, value: Optional[str]) -> None:
        if self.is_mapped_collection:
            self._user_message = value

    @user_message.deleter
    def user_message(self) -> None:
        if self.is_mapped_collection:
            del self._user_message

    _user_message_link = ConfigurationObject.model_property("user_message_link")

    @property
    def user_message_link(self) -> Optional[str]:
        """
        Link to an http or https URL containing additional messaging for
        clients to display to users when interacting with this endpoint.
        """
        if self.is_guest_collection:
            assert self.mapped_collection is not None
            return self.mapped_collection.user_message_link

        return self._user_message_link

    @user_message_link.setter
    def user_message_link(self, value: Optional[str]) -> None:
        if self.is_mapped_collection:
            self._user_message_link = value

    @user_message_link.deleter
    def user_message_link(self) -> None:
        if self.is_mapped_collection:
            del self._user_message_link

    def absolute_root_path(self, username):
        """
        Call :py:code:`StorageGateway.replace_path_vars()` on the root
        path, replacing $HOME, ~, and $USER.

        :param username:
            Connector-specific username to use for $USER.
        :returns:
            A PathRestrictions object or None.
        """
        return StorageGateway.replace_path_vars(
            self.root_path,
            username=username,
            home_directory=self.storage_gateway.home_directory(username),
        )

    def absolute_sharing_restrict_paths(self, username: str) -> PathRestrictions:
        """
        Return the effective sharing path restrictions for the given username,
        with all paths normalized and absolute.

        :param username:
            Connector-specific username of the user to evaluate paths on behalf
            of.
        :returns:
            A PathRestrictions object.
        """

        read = set()  # type: Set[str]
        read_write = set()  # type: Set[str]
        none = set()  # type: Set[str]

        if self.is_mapped_collection:
            policies = self.sharing_policies
            root = self.absolute_root_path(username)
        else:
            policies = self.mapped_collection.sharing_policies
            root = self.mapped_collection.absolute_root_path(username)

        for policy in policies:
            if (not policy.users) or (username in policy.users):
                for path in policy.sharing_restrict_paths.read:
                    read.add(path)
                for path in policy.sharing_restrict_paths.read_write:
                    read_write.add(path)
                for path in policy.sharing_restrict_paths.none:
                    none.add(path)

        home_directory = self.storage_gateway.home_directory(username)

        absolute_paths = PathRestrictions()
        absolute_paths.read = [
            StorageGateway.replace_path_vars(
                re.sub("//*", "/", "/".join([root, r])),
                username=username,
                home_directory=home_directory,
            )
            for r in read
        ]
        absolute_paths.read_write = [
            StorageGateway.replace_path_vars(
                re.sub("//*", "/", "/".join([root, r])),
                username=username,
                home_directory=home_directory,
            )
            for r in read_write
        ]
        absolute_paths.none = [
            StorageGateway.replace_path_vars(
                re.sub("//*", "/", "/".join([root, r])),
                username=username,
                home_directory=home_directory,
            )
            for r in none
        ]
        return absolute_paths

    organization = ConfigurationObject.model_property("organization")
    department = ConfigurationObject.model_property("department")
    keywords = ConfigurationObject.model_property("keywords")
    description = ConfigurationObject.model_property("description")

    @property
    def contact_email(self):
        """
        Mapped collections can inherit contact_email from the endpoint
        """
        value = self.configuration_hash.get("contact_email")
        if value is None and self.is_mapped_collection:
            value = self.endpoint.configuration_hash.get("contact_email")
        return value

    @contact_email.setter
    def contact_email(self, value):
        """
        Collections can set contact_email explicitly. For mapped collections,
        this can override the endpoint's setting.
        """
        if value != self.configuration_hash.get("contact_email"):
            self.configuration_hash["contact_email"] = value

    @contact_email.deleter
    def contact_email(self):
        """
        Collections will allow contact_email to be deleted.
        """
        if "contact_email" in self.configuration_hash:
            del self.configuration_hash["contact_email"]

    contact_info = ConfigurationObject.model_property("contact_info")
    info_link = ConfigurationObject.model_property("info_link")
    scopes = ConfigurationObject.model_property("scopes")

    def check_sharing_user_allowed(self, username: str) -> bool:
        """
        Check whether a user is allowed to create a guest collection, based on a
        mapped collection's policies.

        If the username is in the sharing_users_deny list it is disallowed.
        If the username is in the sharing_users_allow list it is allowed.
        If it is in neither, and sharing_users_allow is defined, it is denied.
        If it is in neither, and sharing_users_allow is not defined, allowed.

        :param username:
            Connector-specific username to check for access.
        :returns:
            True if the user can create a guest collection, False otherwise.
        """
        if self.is_guest_collection or not self.allow_guest_collections:
            return False
        if not self.storage_gateway.check_user_allowed(username):
            return False

        users_allow: Optional[List[str]]
        users_deny: Optional[List[str]]

        try:
            users_allow = self.sharing_users_allow
        except AttributeError:
            users_allow = None
        try:
            users_deny = self.sharing_users_deny
        except AttributeError:
            users_deny = None

        if users_deny and username in users_deny:
            # Explicitly denied
            return False
        if users_allow:
            if username in users_allow:
                # Explicitly allowed
                return True
            else:
                # Others are explicitly allowed, this one is not
                return False
        else:
            # Not denied and nobody explicitly allowed: implicitly allowed
            return True

    @property
    def gcs_private_uuid(self):
        return self.id

    @property
    def gcs_private_parent_endpoint_id(self):
        return self.endpoint.endpoint_id

    @property
    def host_endpoint(self):
        return self.endpoint.endpoint_id

    @property
    def authentication_timeout_mins(self):
        """
        Authentication timeout in minutes for either HA or non-HA
        collections
        """
        return self.storage_gateway.authentication_timeout_mins

    @property
    def require_mfa(self):
        """
        Boolean of whether or not MFA is required for HA collection.s
        """
        return self.storage_gateway.require_mfa

    @staticmethod
    def from_collections_post(
        storage_gateway=None,
        mapped_collection=None,
        domain_name=None,
        endpoint=None,
        user_credential=None,
        authorized_identity=None,
        doc=None,
    ):
        if doc is None or doc.get("DATA_TYPE") not in Collection._DATA_TYPES:
            raise ValueError("Unsupported DATA_TYPE")

        force_encryption = doc.get("force_encryption")
        if storage_gateway.high_assurance:
            force_encryption = True

        if doc["collection_type"] == CollectionType.mapped:
            model_class = Collection.polymorphic_class(storage_gateway.connector_id)

            policies = doc.get("policies", {})

            if policies:
                policies_data_type = policies.get("DATA_TYPE")
                if policies_data_type not in model_class.POLICY_DATA_TYPES:
                    raise ValueError("Unsupported DATA_TYPE")

            if authorized_identity["id"] == current_app.client_id:
                raise ValueError(
                    "Collections must be created by a different identity than "
                    "the endpoint client"
                )
            return model_class(
                doc={
                    **doc,
                    **{
                        "storage_gateway_id": storage_gateway.id,
                        "identity_id": authorized_identity["id"],
                        "domain_name": domain_name,
                        "force_encryption": force_encryption,
                    },
                },
                storage_gateway=storage_gateway,
                endpoint=endpoint,
            )
        else:
            account = storage_gateway.map_identity(authorized_identity)
            allowed = storage_gateway.check_user_allowed(account)
            if not allowed:
                raise errors.ForbiddenError()
            if storage_gateway.require_credential:
                if user_credential is None:
                    user_credential = UserCredential.query.filter(
                        ~UserCredential.deleted,
                        UserCredential.storage_gateway_id == storage_gateway.id,
                        UserCredential.identity_id == authorized_identity["id"],
                        UserCredential.username == account,
                    ).one_or_none()
                if user_credential is None:
                    raise errors.BadRequestError(
                        "A User Credential is needed to create this collection"
                    )

            if not mapped_collection.check_sharing_user_allowed(account):
                raise errors.InvalidUserError(
                    "The connector-specific user that your identity maps to "
                    "is not allowed to create guest collections on this resource.",
                    usernames=[user_credential.username],
                )

            model_class = Collection.polymorphic_class(storage_gateway.connector_id)

            return model_class(
                doc={
                    **doc,
                    **{
                        "storage_gateway_id": storage_gateway.id,
                        "user_credential_id": user_credential.id
                        if user_credential
                        else None,
                        "identity_id": authorized_identity["id"],
                        "domain_name": domain_name,
                        "force_encryption": force_encryption,
                    },
                },
                storage_gateway=storage_gateway,
                mapped_collection=mapped_collection,
                user_credential=user_credential,
                endpoint=endpoint,
            )

    def register_child_client(self):
        """
        register a child client for the collection,
        set the collection's id to the child client id
        and set the collections scopes
        """
        try:
            client, scopes = child_client.register(self)
        except globus_sdk.AuthAPIError as e:
            message = getattr(e, "message", None)
            code = getattr(e, "code", None)

            if not message:
                message = "Unable to create Auth client for collection"
            current_app.logger.warn(message)

            if code == "INVALID_PARAMETERS":
                raise errors.BadRequestError(
                    "Storage Gateway {} is misconfigured: {}".format(
                        self.storage_gateway_id, message
                    )
                )
            else:
                raise errors.BadRequestError(message)

        return client["id"]

    def update_child_client(self):
        """
        Update the child client. Call this when the domain name
        changes.
        """
        if self._domain_name != self.id + "-[DELETED]":
            try:
                return child_client.update_child_client(self)
            except globus_sdk.AuthAPIError as e:
                message = getattr(e, "message", None)
                code = getattr(e, "code", None)

                if not message:
                    message = "Unable to update Auth client for collection"
                current_app.logger.warn(message)

                if code == "INVALID_PARAMETERS":
                    raise errors.BadRequestError(
                        "Storage Gateway {} is misconfigured: {}".format(
                            self.storage_gateway_id, message
                        )
                    )
                else:
                    raise errors.BadRequestError(message)

    def unregister_child_client(self):
        """
        unregister the child client for the collection
        """
        try:
            child_client.unregister(self.id)
        except globus_sdk.exc.GlobusAPIError as e:
            if e.http_status != 404:
                raise

    def create_transfer_endpoint(self):
        try:
            if self.is_mapped_collection:
                create_ep_doc = mapped_collection_create_schema.dump(self)
                current_app.logger.debug(
                    "creating endpoint %s with %s", self.id, create_ep_doc
                )
                return current_app.transfer_client.create_endpoint(create_ep_doc)
            else:
                create_ep_doc = guest_collection_create_schema.dump(self)
                current_app.logger.debug(
                    "creating shared endpoint %s with %s", self.id, create_ep_doc
                )
                return current_app.transfer_client.create_shared_endpoint(create_ep_doc)
        except globus_sdk.TransferAPIError as e:
            log_transfer_exception(e)

            # Special-casing HA collection create with non-HA subscription.
            ha_sub_err = (
                "The endpoint's subscription level does not allow high-assurance"
            )
            message = getattr(e, "message", None)
            code = getattr(e, "code", None)
            if ha_sub_err == message:
                raise errors.SubscriptionRequiredError(
                    ha_sub_err,
                    subscription_level="standard",
                    add_ons=["high_assurance"],
                )
            elif code == "LimitExceeded":
                raise errors.LimitExceededError(
                    error_description="This endpoint has reached the maximum number of collections"
                )
            else:
                raise errors.BadGatewayError(error_description=message)

    def update_scopes(self) -> None:
        current_app.logger.debug("Updating scopes for %s", self)
        data_access_scopes = {
            self.id: None,
            self.domain_name: None,
        }

        auth_client = current_app.auth_client
        if self.is_mapped_collection:
            # Mapped collections have a data_access_scope, which includes text
            # related to the dns name, collection name, and allowed_domains. When
            # any of those are updated, we need to update that scope
            params = {
                "scope_strings": ",".join(
                    [
                        GCSCollectionScopeBuilder(midfix).data_access
                        for midfix in [self.id, self.domain_name]
                    ]
                )
            }
            scopes = auth_client.get("/v2/api/scopes", query_params=params)
            if not scopes.data["scopes"]:
                # Missing, maybe from an old version of GCS?
                resource = "{}/{}/scopes".format(
                    child_client.CLIENTS_API_BASE_URL, self.id
                )
                data_access_res = auth_client.post(
                    resource, data=child_client._data_access_scope(self, self.id)
                )
                for data_access_scope in data_access_res.data["scopes"]:
                    scope_parts = data_access_scope["scope_string"].rsplit("/")[-2:]
                    if scope_parts[1] == "data_access":
                        data_access_scopes[scope_parts[0]] = data_access_scope["id"]
            else:
                for scope in scopes.data["scopes"]:
                    current_app.logger.debug("Updating scope %s", scope["id"])
                    updated_scope = child_client._data_access_scope(self, self.id)
                    current_app.logger.debug("Updating scope to %s", updated_scope)
                    data_access_res = auth_client.put(
                        f"/v2/api/scopes/{scope['id']}", data=updated_scope
                    )
                    data_access_scope = data_access_res.data["scope"]
                    scope_parts = data_access_scope["scope_string"].rsplit("/")[-2:]
                    if scope_parts[1] == "data_access":
                        data_access_scopes[scope_parts[0]] = data_access_scope["id"]

        # The https scope is used by the webapp to access the collection via
        # https. For a mapped collection, This includes collection name and
        # domain in descriptive text
        params = {
            "scope_strings": ",".join(
                [
                    GCSCollectionScopeBuilder(midfix).url_scope_string("https")
                    for midfix in [self.id, self.domain_name]
                ]
            )
        }
        scopes = auth_client.get("/v2/api/scopes", query_params=params)
        for https_scope in scopes.data["scopes"]:
            scope_parts = https_scope["scope_string"].rsplit("/")[-2:]
            updated_scope = child_client._https_scope(self, self.id)
            auth_client.put(f"/v2/api/scopes/{https_scope['id']}", data=updated_scope)

    def update_transfer_endpoint(self):
        if self.is_mapped_collection:
            ep_doc = mapped_collection_update_schema.dump(self)
        else:
            ep_doc = guest_collection_update_schema.dump(self)

        try:
            current_app.logger.debug("updating endpoint %s with %s", self.id, ep_doc)
            return current_app.transfer_client.update_endpoint(self.id, ep_doc)
        except globus_sdk.TransferAPIError as e:
            log_transfer_exception(e)
            if e.http_status == 404:
                self.deleted = True
            else:
                raise errors.BadGatewayError(e.message)

    def __set_owner_string(
        self, owner_id: str, token: str
    ) -> globus_sdk.GlobusHTTPResponse:
        payload = gcs_owner_string_schema.dump(
            dict(owner_id=owner_id, owner_token=token)
        )

        current_app.logger.debug(f"Setting owner_id to {owner_id}")
        try:
            return current_app.transfer_client.put(
                f"/private/endpoint/{self.id}/gcs_owner_string",
                data=payload,
            )
        except globus_sdk.TransferAPIError as e:
            log_transfer_exception(e)
            if e.http_status == 403:
                raise errors.ForbiddenError(e.message)
            raise errors.BadGatewayError(e.message)

    def set_owner_string(self, owner_id: str) -> globus_sdk.GlobusHTTPResponse:
        token = g.auth_state.dependent_token(
            required_scope="urn:globus:auth:scope:transfer.api.globus.org:all"
        )
        return self.__set_owner_string(owner_id, token)

    def reset_owner_string(self) -> globus_sdk.GlobusHTTPResponse:
        client = current_app.transfer_client
        owner_id = current_app.client_id
        authorizer = client.authorizer
        authorizer.ensure_valid_token()
        token = client.authorizer.access_token

        return self.__set_owner_string(owner_id, token)

    def migrate_collection_uuid(
        self,
        persistent_id: str,
        domain_name: str,
        mapped_collection: Optional["Collection"] = None,
    ) -> "Collection":
        assert (
            self.id != persistent_id
        ), "Calling migrate_collection_uuid with a model already using that id"

        # Create a new appsync object data blob that contains the configuration
        # from this collection, then replace the id with the persistent ID
        model_data = self.configuration_schema.dump(self)
        model_data["id"] = persistent_id

        if self.is_guest_collection:
            # Update mapped collection ID to reflect the ID of the mapped
            # collection (which has already had its ID migrated)
            assert mapped_collection is not None
            model_data["mapped_collection_id"] = mapped_collection.id
        else:
            # Update sharing policies in the data blob so that they have new
            # IDs and refer to this collection
            for policy in model_data["sharing_policies"]:
                if policy["id"] == SharingPolicy.default_collection_policy(self.id):
                    policy["id"] = SharingPolicy.default_collection_policy(
                        persistent_id
                    )
                else:
                    policy["id"] = str(uuid.uuid4())
                policy["collection_id"] = persistent_id

        # Use the new domain for this collection to avoid unique constraint
        # problems in the database.
        model_data["domain_name"] = domain_name

        # Construct a new model that includes the configuration from this model
        # but with the ID persistent_id
        new_model = type(self)(
            appsync_data={
                "id": persistent_id,
                "tag": ConfigurationObjectTag.COLLECTION,
                "configuration_hash": model_data,
            },
            storage_gateway=self.storage_gateway,
            mapped_collection=getattr(self, "mapped_collection", None),
            user_credential=getattr(self, "user_credential", None),
            endpoint=getattr(self, "endpoint", None),
        )

        # Force the domain name update in the new model, to propagate the
        # changes to transfer and auth
        new_model.update_domain_name(domain_name, force=True)

        return new_model

    def delete_transfer_endpoint(self, endpoint_id=None):
        if endpoint_id is None:
            endpoint_id = self.id
        try:
            res = current_app.transfer_client.delete_endpoint(endpoint_id)
            return res
        except globus_sdk.TransferAPIError as e:
            log_transfer_exception(e)
            if e.http_status != 404:
                raise errors.BadGatewayError(e.message)

    @run_as_user(user="gcsweb")
    def setup_vhost(self):
        if self.has_domain:
            try:
                self.domain.write_config()
            except Exception as e:
                current_app.logger.error(
                    "Unable to update GCS Manager virtual host: %s", str(e)
                )

        manager_client_secret = current_app.config["CLIENT_SECRET"]
        environment = current_app.environment

        auth_endpoint = globus_sdk.config.get_service_url("auth", environment)
        transfer_endpoint = globus_sdk.config.get_service_url("transfer", environment)

        vhost_conf_dir = os.path.join(current_app.config["HOME"], "etc/httpd/conf.d")
        vhost_conf_file = os.path.join(vhost_conf_dir, "{}.conf".format(self.id))
        vhost_conf_file_tmp = vhost_conf_file + ".gcs.tmp"

        current_app.logger.debug("Writing config to {}".format(vhost_conf_file))

        if not os.path.exists(vhost_conf_dir):
            os.makedirs(vhost_conf_dir, 0o700)

        with open(
            os.open(vhost_conf_file_tmp, os.O_CREAT | os.O_WRONLY | os.O_TRUNC, 0o600),
            "w",
        ) as f:
            f.write(
                render_template(
                    "vhost.conf.jinja2",
                    collection=self,
                    auth_endpoint=auth_endpoint,
                    transfer_endpoint=transfer_endpoint,
                    config_version=__version__,
                    secret=manager_client_secret,
                    subscription_id=current_app.subscription,
                    endpoint=self.endpoint,
                    enable_https=(
                        current_app.subscription
                        and self.configuration_hash.get("enable_https", True)
                    ),
                )
            )
        os.rename(vhost_conf_file_tmp, vhost_conf_file)

    @run_as_user(user="gcsweb")
    def cleanup_vhost(self):
        vhost_conf_dir = os.path.join(current_app.config["HOME"], "etc/httpd/conf.d")

        vhost_conf_file = os.path.join(vhost_conf_dir, "{}.conf".format(self.id))

        if os.path.exists(vhost_conf_file):
            os.remove(vhost_conf_file)

    def write_config(self, delete=False):
        if delete:
            self.cleanup_vhost()
        else:
            self.setup_vhost()

    def _check_model(self):
        if getattr(self, "disable_verify", False) and getattr(
            self, "force_verify", False
        ):
            raise ValueError("disable_verify and force_verify are mutually exclusive")

    def post(self, **doc):
        DATA_TYPE = doc.get("DATA_TYPE")

        if DATA_TYPE not in self._DATA_TYPES:
            raise ValueError("Unsupported DATA_TYPE")

        collection_properties = self.COLLECTION_PROPERTIES.get(DATA_TYPE, [])
        for p in doc:
            if p in ["DATA_TYPE", "connector_id"]:
                continue
            if p not in collection_properties:
                raise AttributeError("Unsupported property {}".format(p))
        self.check_doc(**doc)
        self._put(**doc)
        self._check_model()

        # The checker skips if collection_base_path is undefined, so we
        # check it explicitly here after we know its been set. Once the
        # collection is created, collection_base_path is not changing, so we
        # only need to do this here.
        if "default_directory" in doc:
            self.check_default_directory(doc["default_directory"])

    def put(self, **doc):
        old_domain = self.domain_name
        old_display_name = self.display_name

        DATA_TYPE = doc.get("DATA_TYPE")
        if DATA_TYPE not in self._DATA_TYPES:
            raise ValueError("Unsupported DATA_TYPE")

        collection_properties = self.COLLECTION_PROPERTIES.get(DATA_TYPE, [])

        for p in doc:
            if p == "policies" or p == "DATA_TYPE":
                continue
            if p not in collection_properties:
                raise ValueError("Unsupported property {}".format(p))

        if doc.get("enable_https") and not current_app.subscription:
            raise errors.SubscriptionRequiredError(
                "The enable_https property may not be set on an unmanaged endpoint",
                subscription_level="starter",
            )
        elif (
            self.is_guest_collection
            and doc.get("enable_https")
            and not self.mapped_collection.enable_https
        ):
            raise errors.BadRequestError(
                "The enable_https property may not be enabled a guest collection "
                "when it is not enabled on its mapped collection"
            )

        assert "domain" not in doc

        self.check_doc(**doc)
        self._put(**doc)
        self._check_model()
        self.update_transfer_endpoint()
        if old_domain != self.domain_name or old_display_name != self.display_name:
            # Needed to update the callback uri and name to match the
            # new domain name
            self.update_child_client()
            self.update_scopes()
        self.refresh_configuration_hash()

    def _put(self, DATA_TYPE=None, **doc):
        if DATA_TYPE not in self.COLLECTION_PROPERTIES:
            collection_properties = self.COLLECTION_PROPERTIES[self._DATA_TYPE]
        else:
            collection_properties = self.COLLECTION_PROPERTIES[DATA_TYPE]

        if "policies" not in self.configuration_hash:
            self.configuration_hash["policies"] = dict()

        for p in collection_properties:
            v = doc.get(p)
            if v is not None:
                setattr(self, p, v)
            elif p == "enable_https" and p not in doc:
                self.enable_https = True
            elif hasattr(self, p) and p not in [
                "root_path",
                "deleted",
                "domain_name",
                "collection_type",
            ]:
                delattr(self, p)
        self.policies = doc.get("policies", None)

    def put_domain(self, domain: DomainDict, force: Optional[bool] = False) -> None:
        """
        Update the domain of this collection to a new custom domain. If this is a
        change in the domain name, or the provided force option is True, then
        also update the endpoint document in transfer.

        :param domain:
            A new custom domain to use for this collection.
        :param force:
            If true, update the endpoint document in transfer, even if there is
            no apparent domain name change.
        """
        old_name = None  # type: Optional[str]
        if self.has_domain:
            old_name = self.domain.domain_name

        self.domain = Domain(doc=domain)

        if old_name != self.domain.domain_name or force:
            self.update_domain_name(domain["domain_name"])

    def delete_domain(self, new_domain: str, force: Optional[bool] = False) -> None:
        """
        Remove the custom domain associated with this collection. If this is a
        change in the domain name, or the provided force option is True, then
        also update the endpoint document in transfer.

        :param new_domain:
            The new value for the domain, in the endpoint's collection subdomain. This
            may be altered to be a subdomain of a mapped collection if the owning
            collection has a custom wildcard domain.
        :param force:
            If true, update the server document in transfer, even if there is
            no apparent domain name change.
        """
        had_domain = self.has_domain
        del self.domain

        if had_domain or force:
            self.update_domain_name(new_domain)

    @property
    def policies(self):
        return self.configuration_hash["policies"]

    @policies.setter
    def policies(self, value):
        if value is None:
            value = {}

        # Verify everything in policies is relevant to this connector
        DATA_TYPE = value.get("DATA_TYPE")
        connector_policies = self.CONNECTOR_POLICIES.get(DATA_TYPE, [])

        for p in value:
            if p == "DATA_TYPE":
                if value[p] not in self.POLICY_DATA_TYPES:
                    raise ValueError()
                continue

            if p not in connector_policies:
                raise ValueError("Unknown policy {}".format(p))

        # Update policies
        for p in connector_policies:
            v = value.get(p)

            if v is not None:
                setattr(self, p, v)
            elif hasattr(self, p):
                delattr(self, p)
            else:
                setattr(self, p, None)

    @policies.deleter
    def policies(self):
        connector_policies = self.CONNECTOR_POLICIES.get(self.POLICY_DATA_TYPE, [])
        for p in connector_policies:
            delattr(self, p)
        self.configuration_hash.changed()

    def get_policy(self, name: str):
        if name in self.policies:
            return self.policies[name]
        raise AttributeError(
            "The {} property is not defined for this collection".format(name)
        )

    def set_policy(self, name: str, value: Any):
        if self.policies.get(name) != value:
            self.configuration_hash["policies"][name] = value
            self.configuration_hash.changed()

    def delete_policy(self, name: str):
        if name in self.policies:
            del self.policies[name]
            self.configuration_hash.changed()

    def patch(self, **doc):
        """
        Apply PATCH semantics to this Collection.
        """
        old_domain = self.domain_name
        old_display_name = self.display_name

        DATA_TYPE = doc.get("DATA_TYPE")
        if DATA_TYPE not in self._DATA_TYPES:
            raise AttributeError("Unsupported DATA_TYPE")

        collection_properties = self.COLLECTION_PROPERTIES.get(DATA_TYPE, [])

        # Verify everything in doc is writable
        for p in doc:
            if p == "policies" or p == "DATA_TYPE":
                continue
            if p not in collection_properties:
                raise AttributeError("Unsupported property {}".format(p))

        if doc.get("enable_https") and not current_app.subscription:
            raise errors.SubscriptionRequiredError(
                "The enable_https property may not be set on an unmanaged endpoint",
                subscription_level="starter",
            )
        elif (
            self.is_guest_collection
            and doc.get("enable_https")
            and not self.mapped_collection.enable_https
        ):
            raise errors.BadRequestError(
                "The enable_https property may not be enabled a guest collection "
                "when it is not enabled on its mapped collection"
            )
        self.check_patch_doc(**doc)
        self._patch(**doc)
        self._check_model()
        self.update_transfer_endpoint()
        if old_domain != self.domain_name or old_display_name != self.display_name:
            # Needed to update the callback uri and name to match the
            # new domain name
            self.update_child_client()
            self.update_scopes()

    def _patch(self, DATA_TYPE=None, policies=None, **doc):
        collection_properties = self.COLLECTION_PROPERTIES.get(DATA_TYPE, [])

        # Update properties
        for p in collection_properties:
            if p in doc:
                v = doc[p]
                if v is not None:
                    setattr(self, p, v)
                elif hasattr(self, p):
                    delattr(self, p)

        if policies:
            POLICY_DATA_TYPE = policies.get("DATA_TYPE")
            connector_policies = self.CONNECTOR_POLICIES.get(POLICY_DATA_TYPE, [])

            # Verify everything in policies is understood by connector
            for p in policies:
                if p == "DATA_TYPE":
                    continue

                if p not in connector_policies:
                    raise ValueError("Unknown policy {}".format(p))

            # Update policies
            for p in connector_policies:
                if p not in policies:
                    continue

                v = policies.get(p)
                if v is not None:
                    setattr(self, p, v)
                elif hasattr(self, p):
                    delattr(self, p)
                else:
                    setattr(self, p, None)
        self.refresh_configuration_hash()

    def delete(self, delete_transfer_endpoint=True):
        # Avoid deleting the transfer endpoint on migrated collections
        if delete_transfer_endpoint:
            self.delete_transfer_endpoint()

        if self.owner_role:
            self.owner_role.primary_role = False
        self.unregister_child_client()

        # Scrub some of the data from the collection when it is deleted.
        # The domain name column must be something no other collection
        # can have (hence the uuid-based non-hostname) so that once a
        # collection is deleted, other collections can potentially reuse
        # that name
        self.domain = None
        self.domain_name = self.id + "-[DELETED]"
        self.display_name = self.display_name + " [DELETED]"
        self.collection_base_path = "/"
        self.default_directory = "/"
        super().delete()

    def rollback(self, data=None):
        """
        Rolls back state to a previous version of the configuration_hash,
        including updating transfer.
        """
        if data:
            self.update(data["items"][0])

        self.init_appsync()
        if not self.is_deleted:
            self.update_transfer_endpoint()
            self.refresh_configuration_hash()

    def update(self, new_state):
        super().update(new_state)

    def check_user_credential(self, credential):
        """
        Connector implementations can check whether this credential can be used
        to access this collection. If not, they must raise an appropriate error,
        and if appropriate mark the credential as invalid (but not add it to
        the session).
        """
        pass

    def create_sharing_policy(
        self, sharing_policy_doc: SharingPolicyLoadDict
    ) -> SharingPolicy:
        sharing_policy = SharingPolicy(doc=sharing_policy_doc, collection=self)
        self.configuration_hash["sharing_policies"] = sharing_policy_schema.dump(
            self.sharing_policies, many=True
        )
        return sharing_policy

    def delete_sharing_policy(self, sharing_policy_id: str) -> None:
        for sp in self.sharing_policies:
            if sp.id == sharing_policy_id:
                self.sharing_policies.remove(sp)
                break

        self.configuration_hash["sharing_policies"] = [
            sp
            for sp in self.configuration_hash["sharing_policies"]
            if sp["id"] != sharing_policy_id
        ]

    @hybrid_property
    def is_mapped_collection(self):
        return self.collection_type == CollectionType.mapped.name

    @hybrid_property
    def is_guest_collection(self):
        return self.collection_type == CollectionType.guest.name

    @hybrid_property
    def is_deleted(self):
        # Return whether this collection or its mapped collection is deleted.
        # this allows us to to see guest as deleted before the changes have been
        # propagated to transfer
        return (
            self.storage_gateway.deleted
            or self.deleted
            or (
                self.is_guest_collection
                and self.mapped_collection
                and self.mapped_collection.deleted
            )
        )

    @is_deleted.expression  # type: ignore
    def is_deleted(cls):
        mc = aliased(Collection)
        sg1 = aliased(StorageGateway)
        sg2 = aliased(StorageGateway)
        return case(
            [
                (
                    cls.collection_type == "mapped",
                    select([or_(cls.deleted, sg1.deleted)])
                    .where(cls.storage_gateway_id == sg1.id)
                    .as_scalar(),
                ),
                (
                    cls.collection_type == "guest",
                    select([or_(cls.deleted, mc.deleted, sg2.deleted)])
                    .where(
                        and_(
                            cls.mapped_collection_id == mc.id,
                            cls.storage_gateway_id == sg2.id,
                        )
                    )
                    .as_scalar(),
                ),
            ]
        )

    @property
    def last_accessed_time(self) -> datetime.datetime:
        if self.last_access is not None:
            return datetime.datetime.combine(
                self.last_access, datetime.time(0), tzinfo=datetime.timezone.utc
            )
        raise AttributeError("last_accessed_time")

    @staticmethod
    def __validate_string_list(name: str, value: Optional[List[str]]):
        if value:
            if not isinstance(value, Sequence):
                raise ValueError()
            for v in value:
                if not isinstance(v, str):
                    raise ValueError()

    @property
    def domain(self) -> Domain:
        if self._domain is None:
            value = self.configuration_hash.get("domain")
            if value:
                self._domain = Domain(appsync_data=value)
            else:
                raise AttributeError(
                    "The domain property is not defined on this collection."
                )
        return self._domain

    @domain.setter
    def domain(self, value: Optional[Domain]) -> None:
        if value != self.configuration_hash.get("domain"):
            if value:
                self.configuration_hash["domain"] = value.configuration_schema.dump(
                    value
                )
            else:
                self.configuration_hash.pop("domain", None)
            self._domain = None

    @domain.deleter
    def domain(self):
        if self.configuration_hash.get("domain"):
            self.configuration_hash.pop("domain", None)
            self._domain = None

    @property
    def certificate(self) -> Union[LetsEncryptCertificate, Domain]:
        """
        Reference to an object that has private_key_path, certificate_path, and
        (maybe) certificate_chain_path properties
        """
        if self.is_guest_collection:
            assert self.mapped_collection is not None

        if self.has_domain:
            return self.domain
        elif self.is_guest_collection and self.mapped_collection.has_wildcard_domain:
            return self.mapped_collection.domain
        elif self.endpoint.has_wildcard_domain:
            return self.endpoint.domain
        else:
            return self.endpoint.certificate

    @property
    def has_domain(self):
        return bool(self.configuration_hash.get("domain", False))

    @property
    def has_wildcard_domain(self):
        return bool(self.has_domain and self.domain.wildcard)

    @property
    def _default_sharing_policy_id(self) -> str:
        return SharingPolicy.default_collection_policy(self.id)

    @property
    def oidc_scope(self) -> str:
        """
        OIDCScope value for the mod_auth_openidc configuration for this
        collection's vhost
        """
        scopes: List[str]
        if self.is_guest_collection:
            scopes = [
                TransferScopes.all,
                AuthScopes.openid,
                AuthScopes.profile,
            ]
        elif self.high_assurance:
            scopes = [
                AuthScopes.openid,
                AuthScopes.profile,
            ]
        else:
            scopes = [
                GCSCollectionScopeBuilder(self.id).data_access,
                AuthScopes.openid,
                AuthScopes.profile,
            ]
        return " ".join(scopes)

    def __repr__(self):
        return "<Collection {}>".format(self.id)
