from typing import List, Union, Dict
import uuid

from sqlalchemy import or_, select
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import aliased

from globus.manager.constants import ConfigurationObjectTag
from .account import Account
from . import db, ConfigurationObject
from .storage_gateway import StorageGateway
from .decorators import DocumentChecker, checks_property
from globus.manager.api.schemas.api import user_credential_private_schema


class UserCredential(ConfigurationObject, DocumentChecker):
    """
    User Credential model

    This is the baseline definition of user credentials gateways without any
    connector-specific policies (those are implemented in subclasses). This
    model is uses joined table inheritance with the configuration_object table
    to implement encryption of confidential data and sharing of the data
    across DTNs via AppSync.

    To dynamically resolve the connector ID to the proper concrete model
    implementation, the SQLAlchemy mapper uses the 'secondary_discriminator'
    from the base configuration_object class. This is set to the result of
    concatenating the strings 'STORAGE_GATEWAY' and the hex representation of
    the connector UUID.
    """

    __tablename__ = "user_credential"
    __table_args__ = (
        db.UniqueConstraint("storage_gateway_id", "username", "identity_id"),
    )
    _DATA_TYPE = "user_credential#1.0.0"

    CONNECTOR_POLICIES: Dict[str, List[str]] = {}
    """
    Connector-specific mapping of DATA_TYPE to list of properties which can
    be present in the policy document
    """
    configuration_schema = user_credential_private_schema

    user_credential_id = db.Column(
        "id",
        db.TEXT,
        db.ForeignKey("configuration_object.id"),
        primary_key=True,
        doc="ID of this user credential",
    )

    _identity_id = db.Column(
        "identity_id",
        db.TEXT,
        nullable=False,
        index=True,
        doc="Globus Auth identity which this credential is associated with",
    )

    __username = db.Column("username", db.TEXT, nullable=False, index=True)

    invalid = db.Column(
        db.Boolean,
        nullable=False,
        index=True,
        server_default="0",
        doc="Flag indicating that this credential is no longer valid",
    )

    _storage_gateway_id = db.Column(
        "storage_gateway_id",
        db.TEXT,
        db.ForeignKey("storage_gateway.id"),
        nullable=False,
        doc="Storage gateway this credential is associated with",
    )

    storage_gateway: StorageGateway
    storage_gateway = db.relationship(
        "StorageGateway",
        foreign_keys=[_storage_gateway_id],
        uselist=False,
        lazy="joined",
    )

    def __init__(self, appsync_data=None, doc=None, storage_gateway=None):
        if appsync_data is not None:
            args = {k: v for k, v in appsync_data.items()}
            args["tag"] = ConfigurationObjectTag.USER_CREDENTIAL
            args["secondary_discriminator"] = self.connector_id
            if storage_gateway:
                self.storage_gateway = storage_gateway
            super().__init__(**args)
            self.init_appsync()
        elif doc:
            if doc.get("DATA_TYPE") == "user_credential#1.0.0":
                # Use a type 5 uuid based on the unique constraints so that
                # collisions in appsync will happen if there would be
                # collisions in the database
                user_credential_id = self.id_for_account(Account(**doc))
                self.storage_gateway = storage_gateway
                self.username = doc.get("username")

                super().__init__(
                    id=user_credential_id,
                    tag=ConfigurationObjectTag.USER_CREDENTIAL,
                    secondary_discriminator=self.connector_id,
                )
                if "provisioned" in doc:
                    self.provisioned = doc["provisioned"]

                self.put(**{**doc, "id": user_credential_id})
                self.refresh_configuration_hash()
            else:
                raise ValueError("Unsupported DATA_TYPE")
        else:
            raise ValueError("Invalid invocation of UserCredential constructor.")

    @staticmethod
    def id_for_account(account: Account) -> uuid.UUID:
        """
        Compute the ID for the UserCredential for this particular account
        mapping. This is used to produce consistent UUIDs for accounts that
        meet the same uniqueness constraints in the local GCS database.

        :param account:
            The account information to use to compute this information.

        :returns:
            Returns a UUID object containing the unique identifier for this
            account.
        """
        UUID_NAMESPACE = uuid.UUID("1d2571e0-03c0-4657-88fa-4699ba3901a7")
        uuid_name = ":".join(
            [account.storage_gateway_id, account.username, account.identity_id]
        )
        return uuid.uuid5(UUID_NAMESPACE, uuid_name)

    def init_appsync(self):
        if self.configuration_hash:
            self._put(**self.configuration_hash)

    @classmethod
    def polymorphic_class(cls, discriminator):
        return ConfigurationObject.polymorphic_class(
            ConfigurationObjectTag.USER_CREDENTIAL, discriminator
        )

    @property
    def connector_properties(self) -> List[str]:
        return []

    @checks_property(prop="connector_id")
    def check_connector_id(self, value: str):
        if value and str(value) != self.connector_id:
            raise AttributeError()

    @property
    def display_name(self) -> str:
        """
        User-friendly name of this credential
        """
        display_name = self.configuration_hash.get("display_name")
        if display_name is None:
            raise AttributeError(
                "The display_name property is not defined for this user credential"
            )
        return display_name

    @display_name.setter  # type: ignore
    def display_name(self, value: str) -> None:
        self.configuration_hash["display_name"] = value

    @hybrid_property
    def username(self) -> str:
        """
        Connector-specific username for this credential
        """
        return self.__username

    @username.setter  # type: ignore
    def username(self, value: str):
        if self.__username != value:
            self.__username = value

    @checks_property(prop="username")
    def _check_username(self, value: str):
        if self.__username and self.__username != value:
            raise AttributeError()

    @hybrid_property
    def identity_id(self) -> str:
        """
        Auth identity ID of the owner of this credential.
        """
        return self._identity_id

    @identity_id.setter  # type: ignore
    def identity_id(self, value: Union[str, uuid.UUID]):
        if not self._identity_id and value:
            self._identity_id = str(value)

    @checks_property(prop="identity_id")
    def _check_identity_id(self, value: Union[str, uuid.UUID]):
        if self._identity_id and str(value) != self._identity_id:
            raise AttributeError()

    @hybrid_property
    def storage_gateway_id(self) -> str:
        """
        Identity of StorageGateway that this UserCredential can be used with.
        """
        return self._storage_gateway_id

    @storage_gateway_id.setter  # type: ignore
    def storage_gateway_id(self, value: Union[str, uuid.UUID]):
        if not self._storage_gateway_id and value:
            self._storage_gateway_id = str(value)

    @checks_property(prop="storage_gateway_id")
    def _check_storage_gateway_id(self, value: Union[str, uuid.UUID]):
        if self._storage_gateway_id and str(value) != self.storage_gateway_id:
            raise AttributeError()

    @property
    def home_directory(self) -> str:
        return self.storage_gateway.home_directory(self.username)

    @property
    def provisioned(self) -> bool:
        return bool(self.configuration_hash.get("provisioned", True))

    @provisioned.setter
    def provisioned(self, value: bool):
        self.configuration_hash["provisioned"] = bool(value)

    def patch(self, **doc):
        if doc["DATA_TYPE"] != self.configuration_schema._DATA_TYPE:
            raise ValueError("Unsupported DATA_TYPE")

        self.check_patch_doc(**doc)
        self._patch(**doc)
        self.refresh_configuration_hash()

    def _patch(self, DATA_TYPE=None, **doc):
        base_properties = [
            "id",
            "connector_id",
            "identity_id",
            "storage_gateway_id",
            "display_name",
            "username",
            "invalid",
            "provisioned",
            "deleted",
        ]
        for p in base_properties:
            if p in doc:
                setattr(self, p, doc[p])

        policies = doc.get("policies")
        if not policies:
            policies = {"DATA_TYPE": self.POLICY_DATA_TYPE}
        self._patch_policies(**policies)

    def _patch_policies(self, DATA_TYPE=None, **policies):
        DATA_TYPE = policies.get("DATA_TYPE", self.POLICY_DATA_TYPE)
        if DATA_TYPE not in self.POLICY_DATA_TYPES:
            raise ValueError("Invalid DATA_TYPE {}".format(DATA_TYPE))
        _connector_properties = self.CONNECTOR_POLICIES.get(
            DATA_TYPE, self.connector_properties
        )

        for p in policies:
            if p in _connector_properties:
                setattr(self, p, policies[p])
            elif p != "DATA_TYPE":
                raise ValueError("Unknown policy {}".format(p))

    def put(self, **doc):
        try:
            doc["provisioned"] = self.provisioned
            self.check_doc(**doc)
            self._put(**doc)
        finally:
            self.refresh_configuration_hash()

    def _put(self, DATA_TYPE=None, policies=None, **doc):
        if DATA_TYPE != self.configuration_schema._DATA_TYPE:
            raise ValueError("Unsupported DATA_TYPE")

        if doc.get("display_name") is None:
            doc["display_name"] = getattr(self, "display_name", None)

        if doc.get("id") is None:
            doc["id"] = self.id

        base_properties = [
            "id",
            "connector_id",
            "identity_id",
            "storage_gateway_id",
            "display_name",
            "username",
            "invalid",
            "provisioned",
            "deleted",
        ]
        for p in base_properties:
            if p in ["invalid", "deleted"]:
                default_value = False
            elif p == "provisioned":
                default_value = True
            else:
                default_value = None
            setattr(self, p, doc.get(p, default_value))

        if not policies:
            policies = {"DATA_TYPE": self.POLICY_DATA_TYPE}
        self._put_policies(**policies)

    def _put_policies(self, DATA_TYPE=None, **policies):
        DATA_TYPE = policies.get("DATA_TYPE", self.POLICY_DATA_TYPE)
        if DATA_TYPE not in self.POLICY_DATA_TYPES:
            raise ValueError("Invalid DATA_TYPE {}".format(DATA_TYPE))
        _connector_properties = self.CONNECTOR_POLICIES.get(
            DATA_TYPE, self.connector_properties
        )

        for p in policies:
            if p in _connector_properties:
                setattr(self, p, policies.get(p))
            elif p != "DATA_TYPE":
                raise ValueError("Unknown policy {}".format(p))
        for p in _connector_properties:
            if p not in policies:
                delattr(self, p)

    @property
    def policies(self) -> MutableDict:
        if "policies" not in self.configuration_hash:
            self.configuration_hash["policies"] = MutableDict()
        return self.configuration_hash["policies"]

    @hybrid_property
    def is_deleted(self):
        return self.deleted or self.storage_gateway.deleted

    @is_deleted.expression  # type: ignore
    def is_deleted(cls):
        sg = aliased(StorageGateway)
        return (
            select([or_(cls.deleted, sg.deleted)])
            .where(cls.storage_gateway_id == sg.id)
            .as_scalar()
        )

    def delete(self):
        self.invalid = True
        self.provisioned = False
        super().delete()

    def rollback(self, data=None):
        if data:
            self.update(data["items"][0])

    def needs_refresh(self):
        """
        Return True if credential requires some refresh action to be performed
        before the credential can be used. Subclasses should implement this
        if necessary to check for expiring credentials.
        """
        return False

    def refresh(self):
        """
        Update an expired temporary credential. Subclasses should implement
        this if necessary to refresh credentials.
        """
        pass


__all__ = ["UserCredential"]
