from typing import Dict

import marshmallow as ma
from marshmallow_oneofschema import OneOfSchema

from globus.manager.constants import CollectionType
from .data_type import DataTypeField, DataTypeSchema, data_type_schema
from .domain import domain_schema, domain_private_schema
from ..patch_schema import create_patch_schema
from ..cleanup_uuid_mixin import CleanupUUIDMixin
from ..enum_field import EnumField
from .path_restrictions import path_restrictions_schema
from .schema_version_checker import version_safe_type_schemas
from .sharing_policy import sharing_policy_schema


class CollectionPolicies(ma.fields.Method):
    def _jsonschema_type_mapping(self):
        type_mapping = {
            "oneOf": [],
        }
        refs = type_mapping["oneOf"]
        for policy in CollectionSchema_1_0_0.connector_policies_schemas.values():
            title = getattr(policy, "title", policy.__class__.__name__)
            refs.append({"$ref": f"#/components/schemas/{title}"})

        metadata = self.metadata.get("metadata", {})
        metadata.update(self.metadata)
        for md_key, md_val in metadata.items():
            if md_key in ("metadata", "name"):
                continue
            type_mapping[md_key] = md_val
        return type_mapping


class CollectionSchema_1_0_0(ma.Schema, CleanupUUIDMixin):
    __doc__ = """
    A collection consists of metadata about the collection, a DNS
    domain for accessing data on the collection, and configuration on
    the Data Transfer Nodes to access the collection data. Globus
    Connect Server version 5 supports two types of collections:
    **mapped** and **guest**.
    """
    _DATA_TYPE = "collection#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    id = ma.fields.UUID(
        dump_only=True,
        description="""
        Unique identifier for this collection. This is assigned
        by the GCS manager when creating a collection.
        """,
    )
    domain_name = ma.fields.String(
        description="""
        DNS name of the virtual host serving this collection. For mapped
        collections with do not have a custom domain, this may be specified as
        part of the input document to create the collection, otherwise this is
        a read-only property. When included in the input, the name is
        restricted to be a subdomain of the endpoint, and the input name label
        may not start with `m-` or `g-`.
        """
    )
    connector_id = ma.fields.UUID(
        dump_only=True,
        description="Id of the connector type that is used by this collection.",
    )
    manager_url = ma.fields.Url(
        schemes=["https"],
        dump_only=True,
        description="""
        URL of the GCS Manager API service for the endpoint hosting this
        collection.
        """,
    )
    https_url = ma.fields.Url(
        schemes=["https"],
        dump_only=True,
        description="HTTPS URL for the data on this collection.",
    )
    tlsftp_url = ma.fields.Url(
        schemes=["tlsftp"],
        dump_only=True,
        description="TLSFTP URL for the data on this collection.",
    )
    collection_type = EnumField(
        CollectionType,
        required=True,
        description="""
        Type of collection. A `mapped` collection requires an account on
        the system to access the administrator-defined collection. A
        `guest` collection allows users to share access to their data on a
        Storage Gateway by registering a credential with the GCS Manager.
        """,
    )
    display_name = ma.fields.String(
        required=True,
        validate=[ma.validate.Length(max=128), ma.validate.Regexp(r"\S")],
        description="""
        Friendly name for the collection. Unicode string, max 128
        characters, no new lines (`\\r` or `\\n`).
        """,
    )
    identity_id = ma.fields.UUID(
        required=False,
        description="""
        Globus Auth identity to who acts as the owner of this collection.
        This identity is an `administrator` on the collection.
        """,
    )
    storage_gateway_id = ma.fields.UUID(
        description="""
        Unique ID of the Storage Gateway which this collection provides
        access to. This value can not change after the collection is
        created.
        """
    )
    mapped_collection_id = ma.fields.UUID(
        required=False,
        description="""
        Unique ID of the Mapped Collection which this guest collection is
        associated with. This is set on creation and may not be changed.
        For a Guest Collection, this must be set, and policies related
        sharing (`allow_guest_collections`, `sharing_restrict_paths`) will
        always reflect the values in the Mapped Collection definition and
        may not be changed on this Guest Collection.
        """,
    )
    root_path = ma.fields.String(
        required=False,
        dump_only=True,
        description="""
        Absolute root path of the collection. All data access
        is done relative to this path. On a guest collection,
        this value is only visible if the caller has an
        administrator role on both the guest collection and the
        mapped collection it is created on.
        """,
    )
    collection_base_path = ma.fields.String(
        required=True,
        validate=[ma.validate.Regexp(r"^(/|\$HOME)")],
        description="""
        Path to be interpreted as the base path when creating a new
        collection. It is interpreted differently depending on the
        collection type being created. For a mapped collection, this is an
        absolute path on the storage system named by the
        storage_gateway_id.  For a guest collection, this is a relative
        path relative to the value of the `root_path` attribute on the
        mapped collection with the same Id as the `mapped_collection_id`
        property.  This may not be changed once the collection is created.
        """,
    )
    default_directory = ma.fields.String(
        validate=[ma.validate.Regexp(r"^(/|\$HOME|~)")],
        description="""
        Default directory when accessing the collection. This may include
        the special string `$USER` which is evaluated at access time to be
        the connector-specific username accessing the data.

        If the collection is mapped collection with a
        **collection_base_path** value of `/`, this value can also begin
        with the values `/~/` and `$HOME`, which are replaced by the user's
        home directory, or `/` if the connector does not support the
        concept of a home directory.
        """,
    )
    public = ma.fields.Boolean(
        required=True,
        description="""
        Flag indicating whether this collection is visible to other Globus
        users.
        """,
    )
    force_encryption = ma.fields.Boolean(
        description="""
        Flag indicating whether all data transfers to and from this
        collection are always encrypted.

        __New in v5.4.17__: If a mapped collection forces encryption, all
        of its guest collections must as well.  If this option is used on a
        mapped collection, the value is propagated to its guest
        collections.
        """
    )

    disable_verify = ma.fields.Boolean(
        description="""
        Flag indicating that this endpoint does not support computing
        checksums, needed for the verify_checksum option of transfer.
        """
    )
    user_credential_id = ma.fields.UUID(
        required=False,
        description="""
        The ID of the User Credential which is used to access data on this
        collection. This credential must be owned by the collection's
        identity_id.
        """,
    )
    organization = ma.fields.String(
        validate=[ma.validate.Length(max=1024)],
        required=False,
        description="""
        Organization that runs the server(s) represented by the endpoint.
        Optional to preserve backward compatibility, but will eventually be
        required and all clients are encouraged to require users to specify
        it.  Unicode string, max 1024 characters, no new lines.
        """,
    )
    department = ma.fields.String(
        validate=[ma.validate.Length(max=1024)],
        allow_none=True,
        required=False,
        description="""
        Department within organization that runs the server(s).
        Searchable. Optional. Unicode string, max 1024
        characters, no new lines.
        """,
    )
    keywords = ma.fields.List(
        ma.fields.String(),
        required=False,
        description="""
        List of search keywords for the
        endpoint.  Optional. Unicode string, max 1024
        characters total across all strings.
        """,
    )
    description = ma.fields.String(
        validate=[ma.validate.Length(max=4096)],
        allow_none=True,
        required=False,
        description="A description of the collection.",
    )
    contact_email = ma.fields.Email(
        allow_none=True,
        required=False,
        description="Email address of the support contact for the collection.",
    )
    contact_info = ma.fields.String(
        validate=[ma.validate.Length(max=4096)],
        allow_none=True,
        description="""
        Other non-email contact information for the collection, e.g.  phone
        and mailing address.
        """,
    )
    info_link = ma.fields.Url(
        allow_none=True,
        required=False,
        description="Link to a web page with more information about the collection",
    )
    high_assurance = ma.fields.Boolean(
        dump_only=True,
        description="""
        Flag indicating if this collection is created on a high assurance
        Storage Gateway.
        """,
    )
    authentication_timeout_mins = ma.fields.Int(
        dump_only=True,
        description="""
        Timeout (in minutes) during which a user is required to have
        authenticated in a session to access this storage gateway.
        """,
    )
    allow_guest_collections = ma.fields.Boolean(
        required=False,
        description="""
        Flag indicating if this Collection allows users to create guest
        collections on it. This is always false if this is a guest
        collection. If this is changed to false on a mapped collection with
        associated guest collections, those collections will no longer be
        accessible.
        """,
    )
    policies = CollectionPolicies(
        serialize="dump_policies",
        deserialize="load_policies",
        description="Connector-specific collection policies",
    )
    sharing_restrict_paths = ma.fields.Nested(
        path_restrictions_schema,
        required=False,
        allow_none=True,
        description="""
        Restrictions on which paths may be shared in guest collections related
        to this mapped collection. On the mapped collection, these paths are
        relative to the root_path property of the mapped collection. On a guest
        collection, they are absolute paths from the storage root.
        """,
    )
    deleted = ma.fields.Boolean(
        required=False,
        description="Flag indicating that this collection has been deleted",
    )

    private = [
        "root_path",
        "collection_base_path",
        "default_directory",
        "sharing_restrict_paths",
        "deleted",
    ]
    uuid_fields = [
        "id",
        "identity_id",
        "storage_gateway_id",
        "user_credential_id",
        "mapped_collection_id",
    ]

    @ma.post_load
    def clean_keywords(self, data, **kwargs):
        res = data.copy()
        keywords = res.get("keywords")
        if keywords:
            res["keywords"] = [kw.strip() for kw in data["keywords"]]
        return res

    @ma.validates("keywords")
    def validate_keywords(self, data, **kwargs):
        if data:
            kw = ",".join(data)
            if len(kw) > 1024:
                raise ma.ValidationError("Longer than maximum length 1024.")

    @ma.validates_schema
    def validate_mapped_and_guest_only_fields(self, data, **kwargs):
        # Patch may not have collection type set, so we can't fully validate
        # without loading the object from the database
        collection_type = data.get("collection_type")
        if collection_type == CollectionType.mapped:
            field = None
            if "user_credential_id" in data:
                field = "user_credential_id"
                raise ma.ValidationError(
                    "Invalid field for mapped collection.", field_name=field
                )
        elif collection_type == CollectionType.guest:
            for field in ["allow_guest_collections", "sharing_restrict_paths"]:
                if field in data and data[field]:
                    raise ma.ValidationError(
                        "Invalid field for guest collection.", field_name=field
                    )

    def load_policies(self, value):
        return data_type_schema.load(value, partial=self.partial)

    def dump_policies(self, obj):
        connector_id = str(ma.utils.get_value(obj, "connector_id"))
        if not self.load_only:
            schema = self.connector_policies_private_schemas.get(connector_id)
        else:
            schema = self.connector_policies_schemas.get(connector_id)

        if schema:
            return schema.dump(obj)
        else:
            raise ValueError("Unable to serialize {} {}".format(obj, connector_id))

    # connectors may register their schema types in this dict, with
    # the connector ID as the key
    connector_policies_schemas = {}  # type: Dict[str, ma.Schema]
    connector_policies_private_schemas = {}  # type: Dict[str, ma.Schema]

    @classmethod
    def register_schema(cls, public=None, private=None):
        if public:
            connector_id = public.connector_id
            cls.connector_policies_schemas[connector_id] = public
        if private:
            connector_id = private.connector_id
            cls.connector_policies_private_schemas[connector_id] = private


class CollectionSchema_1_1_0(CollectionSchema_1_0_0):
    __doc__ = (
        CollectionSchema_1_0_0.__doc__
        + """
    Version 1.1.0 adds support for enabling or disabling https access for
    individual collections, as well as the ability for collection
    administrators to add an optional message and web link to be shown on
    the Globus web app when users visit the collection.
    """
    )
    _DATA_TYPE = "collection#1.1.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    enable_https = ma.fields.Boolean(
        required=False,
        allow_none=False,
        description="""
        Boolean flag indicating whether this collection should support
        HTTPS. This value can be set on mapped collections or guest
        collections. However, it may not be set to True on a guest
        collection if the value on the related mapped collection is False.
        """,
    )
    user_message = ma.fields.Str(
        required=False,
        allow_none=True,
        validate=[ma.validate.Length(max=64)],
        description="""
        A message for clients to display to users when interacting with
        this collection. For guest collections, this property is read-only
        and is the same as the value of its related mapped collection. The
        message may be up to 64 characters long.
        """,
    )

    user_message_link = ma.fields.Url(
        required=False,
        allow_none=True,
        schemes=["http", "https"],
        description="""
        Link to additional messaging for clients to display to users
        when interacting with this endpoint, linked to an
        HTTP or HTTPS URL. For guest collections, this property is
        read-only and is the same as the value of its related mapped
        collection.
        """,
    )

    @ma.validates_schema
    def validate_mapped_only_1_1_0_fields(self, data, **kwargs):
        # Patch may not have collection type set, so we can't fully validate
        # without loading the object from the database
        collection_type = data.get("collection_type")
        if collection_type == CollectionType.guest:
            for field in ["user_message", "user_message_link"]:
                if field in data and data[field]:
                    raise ma.ValidationError(
                        "Invalid field for guest collection.", field_name=field
                    )


class CollectionSchema_1_2_0(CollectionSchema_1_1_0):
    __doc__ = (
        CollectionSchema_1_1_0.__doc__
        + """
    Version 1.2.0 adds the ability to enable or disable sharing by specific
    users.
    """
    )
    _DATA_TYPE = "collection#1.2.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    sharing_users_allow = ma.fields.List(
        ma.fields.Str,
        required=False,
        allow_none=True,
        description="""
        List of connector-specific usernames allowed to create new guest
        collections on this mapped collection.
        """,
    )
    sharing_users_deny = ma.fields.List(
        ma.fields.Str,
        required=False,
        allow_none=True,
        description="""
        List of connector-specific usernames denied access to
        create new guest collections on this mapped collection.
        """,
    )

    private = CollectionSchema_1_1_0.private + [
        "sharing_users_allow",
        "sharing_users_deny",
    ]

    @ma.validates_schema
    def validate_mapped_only_1_2_0_fields(self, data, **kwargs):
        # Patch may not have collection type set, so we can't fully validate
        # without loading the object from the database
        collection_type = data.get("collection_type")
        if collection_type == CollectionType.guest:
            for field in ["sharing_users_allow", "sharing_users_deny"]:
                if field in data and data[field]:
                    raise ma.ValidationError(
                        "Invalid field for guest collection.", field_name=field
                    )


class CollectionSchema_1_3_0(CollectionSchema_1_2_0):
    __doc__ = (
        CollectionSchema_1_2_0.__doc__
        + """
    Version 1.3.0 add support for custom DNS domains on collections.
    """
    )
    _DATA_TYPE = "collection#1.3.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    domain = ma.fields.Nested(domain_schema, dump_only=True)

    private = CollectionSchema_1_2_0.private + [
        "domain",
    ]


class CollectionSchema_1_4_0(CollectionSchema_1_3_0):
    __doc__ = (
        CollectionSchema_1_3_0.__doc__
        + """
    Version 1.4.0 allows optional multi-factor authentication requirements to
    high assurance collections and the ability to require checksums when
    transferring data on this collection.
    """
    )
    _DATA_TYPE = "collection#1.4.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    require_mfa = ma.fields.Boolean(
        dump_only=True,
        default=False,
        description="""
        Flag indicating if the storage_gateway requires multi-factor
        authentication. Only applies to high assurance storage gateways.
        """,
    )
    force_verify = ma.fields.Boolean(
        description="""
        Flag indicating that this endpoint requires computing checksums,
        needed for the verify_checksum option of transfer.
        """
    )


class CollectionSchema_1_5_0(CollectionSchema_1_4_0):
    __doc__ = (
        CollectionSchema_1_4_0.__doc__
        + """
    Version 1.5.0 allows administrators to disable ACLs that would allow
    anonymous users to have write access to an endpoint.
    """
    )
    _DATA_TYPE = "collection#1.5.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    disable_anonymous_writes = ma.fields.Boolean(
        description="""
        Flag indicating if guest collections on this mapped collection an
        allow anonymous write ACLs or not. This flag is always true for high
        assurance collections. For non-high assurance mapped collections, the
        default value is false.
        """
    )


class CollectionSchema_1_6_0(CollectionSchema_1_5_0):
    __doc__ = (
        CollectionSchema_1_5_0.__doc__
        + """
    Version 1.6.0 allows administrators of mapped collections to associate
    policies that users accessing guest collections must meet beyond the
    guest collection ACLs.
    """
    )
    _DATA_TYPE = "collection#1.6.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    guest_auth_policy_id = ma.fields.UUID(
        allow_none=True,
        description="""
        Authentication policy set on mapped collections and inherited by its
        guest collections. During authorization, the authentication policy must
        be satisfied before ACLs are considered. Read-only on guest
        collections. (**Added in API 1.15.0**)
        """,
    )

    uuid_fields = CollectionSchema_1_5_0.uuid_fields + ["guest_auth_policy_id"]


class CollectionSchema_1_7_0(CollectionSchema_1_6_0):
    __doc__ = (
        CollectionSchema_1_6_0.__doc__
        + """
    Version 1.7.0 increases the maximum allowed length of the user_message
    property.
    """
    )
    _DATA_TYPE = "collection#1.7.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    user_message = ma.fields.Str(
        required=False,
        allow_none=True,
        validate=[ma.validate.Length(max=256)],
        description="""
            A message for clients to display to users when interacting with
            this collection. For guest collections, this property is read-only
            and is the same as the value of its related mapped collection. The
            message may be up to 256 characters long.
        """,
    )


class CollectionSchema_1_8_0(CollectionSchema_1_7_0):
    __doc__ = (
        CollectionSchema_1_7_0.__doc__
        + """
    Version 1.8.0 adds the delete_protected property. While it is set to true
    on a mapped collection, the collection may not be deleted.
    """
    )
    _DATA_TYPE = "collection#1.8.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    delete_protected = ma.fields.Boolean(
        description="""
        If set to true, this collection can not be deleted. This property
        is available only on mapped collections.
        """
    )


class CollectionSchema_1_9_0(CollectionSchema_1_8_0):
    __doc__ = (
        CollectionSchema_1_8_0.__doc__
        + """
    Version 1.9.0 adds the read-only last_access and created_at properties.
    """
    )
    _DATA_TYPE = "collection#1.9.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    created_at_date = ma.fields.Date(
        allow_none=True,
        data_key="created_at",
        dump_only=True,
        description="Date on which this collection was created",
    )
    last_access = ma.fields.Date(
        allow_none=True,
        dump_only=True,
        description="Date on which this collection was last accessed",
    )


class CollectionConfigurationSchema(CollectionSchema_1_9_0):
    """
    Collection schema stored in appsync. Contains the full Domain object,
    including the private key if it is managed.
    """

    domain = ma.fields.Nested(domain_private_schema, dump_only=True)
    sharing_policies = ma.fields.List(
        ma.fields.Nested(sharing_policy_schema), dump_only=True
    )


collection_1_0_0_schema = CollectionSchema_1_0_0(
    load_only=CollectionSchema_1_0_0.private
)
collection_1_0_0_private_schema = CollectionSchema_1_0_0()
collection_1_0_0_patch_schema = create_patch_schema(CollectionSchema_1_0_0)

collection_1_1_0_schema = CollectionSchema_1_1_0(
    load_only=CollectionSchema_1_1_0.private
)
collection_1_1_0_private_schema = CollectionSchema_1_1_0()
collection_1_1_0_patch_schema = create_patch_schema(CollectionSchema_1_1_0)

collection_1_2_0_schema = CollectionSchema_1_2_0(
    load_only=CollectionSchema_1_2_0.private
)
collection_1_2_0_private_schema = CollectionSchema_1_2_0()
collection_1_2_0_patch_schema = create_patch_schema(CollectionSchema_1_2_0)

collection_1_3_0_schema = CollectionSchema_1_3_0(
    load_only=CollectionSchema_1_3_0.private
)
collection_1_3_0_private_schema = CollectionSchema_1_3_0()
collection_1_3_0_patch_schema = create_patch_schema(CollectionSchema_1_3_0)

collection_1_4_0_schema = CollectionSchema_1_4_0(
    load_only=CollectionSchema_1_4_0.private
)
collection_1_4_0_private_schema = CollectionSchema_1_4_0()
collection_1_4_0_patch_schema = create_patch_schema(CollectionSchema_1_4_0)
collection_1_5_0_schema = CollectionSchema_1_5_0(
    load_only=CollectionSchema_1_5_0.private
)
collection_1_5_0_private_schema = CollectionSchema_1_5_0()
collection_1_5_0_patch_schema = create_patch_schema(CollectionSchema_1_5_0)
collection_1_6_0_schema = CollectionSchema_1_6_0(
    load_only=CollectionSchema_1_6_0.private
)
collection_1_6_0_private_schema = CollectionSchema_1_6_0()
collection_1_6_0_patch_schema = create_patch_schema(CollectionSchema_1_6_0)
collection_1_7_0_schema = CollectionSchema_1_7_0(
    load_only=CollectionSchema_1_7_0.private
)
collection_1_7_0_private_schema = CollectionSchema_1_7_0()
collection_1_7_0_patch_schema = create_patch_schema(CollectionSchema_1_7_0)
collection_1_8_0_schema = CollectionSchema_1_8_0(
    load_only=CollectionSchema_1_8_0.private
)
collection_1_8_0_private_schema = CollectionSchema_1_8_0()
collection_1_8_0_patch_schema = create_patch_schema(CollectionSchema_1_8_0)
collection_configuration_schema = CollectionConfigurationSchema()
collection_1_9_0_schema = CollectionSchema_1_9_0(
    load_only=CollectionSchema_1_9_0.private
)
collection_1_9_0_private_schema = CollectionSchema_1_9_0()
collection_1_9_0_patch_schema = create_patch_schema(CollectionSchema_1_9_0)
collection_configuration_schema = CollectionConfigurationSchema()


class Collection(OneOfSchema):
    __doc__ = CollectionSchema_1_9_0.__doc__

    _DATA_TYPE = collection_1_9_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = version_safe_type_schemas(
        {
            "collection#1.0.0": collection_1_0_0_schema,
            "collection#1.1.0": collection_1_1_0_schema,
            "collection#1.2.0": collection_1_2_0_schema,
            "collection#1.3.0": collection_1_3_0_schema,
            "collection#1.4.0": collection_1_4_0_schema,
            "collection#1.5.0": collection_1_5_0_schema,
            "collection#1.6.0": collection_1_6_0_schema,
            "collection#1.7.0": collection_1_7_0_schema,
            "collection#1.8.0": collection_1_8_0_schema,
            "collection#1.9.0": collection_1_9_0_schema,
        }
    )
    type_field_remove = False

    def get_obj_type(self, obj):
        return self._DATA_TYPE

    @classmethod
    def register_schema(cls, public=None, private=None):
        for c in cls.type_schemas.values():
            c.register_schema(public=public, private=private)


class CollectionPrivate(OneOfSchema):
    __doc__ = CollectionSchema_1_9_0.__doc__
    _DATA_TYPE = collection_1_9_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = {
        "collection#1.0.0": collection_1_0_0_private_schema,
        "collection#1.1.0": collection_1_1_0_private_schema,
        "collection#1.2.0": collection_1_2_0_private_schema,
        "collection#1.3.0": collection_1_3_0_private_schema,
        "collection#1.4.0": collection_1_4_0_private_schema,
        "collection#1.5.0": collection_1_5_0_private_schema,
        "collection#1.6.0": collection_1_6_0_private_schema,
        "collection#1.7.0": collection_1_7_0_private_schema,
        "collection#1.8.0": collection_1_8_0_private_schema,
        "collection#1.9.0": collection_1_9_0_private_schema,
    }
    type_field_remove = False

    def get_obj_type(self, obj):
        return self._DATA_TYPE


class CollectionPatch(OneOfSchema):
    __doc__ = CollectionSchema_1_9_0.__doc__
    _DATA_TYPE = collection_1_9_0_patch_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = version_safe_type_schemas(
        {
            "collection#1.0.0": collection_1_0_0_patch_schema,
            "collection#1.1.0": collection_1_1_0_patch_schema,
            "collection#1.2.0": collection_1_2_0_patch_schema,
            "collection#1.3.0": collection_1_3_0_patch_schema,
            "collection#1.4.0": collection_1_4_0_patch_schema,
            "collection#1.5.0": collection_1_5_0_patch_schema,
            "collection#1.6.0": collection_1_6_0_patch_schema,
            "collection#1.7.0": collection_1_7_0_patch_schema,
            "collection#1.8.0": collection_1_8_0_patch_schema,
            "collection#1.9.0": collection_1_9_0_patch_schema,
        }
    )
    type_field_remove = False

    def get_obj_type(self, obj):
        return self._DATA_TYPE


collection_schema = Collection(load_only=CollectionSchema_1_8_0.private)
CollectionSchema = Collection
collection_private_schema = CollectionPrivate()
collection_patch_schema = CollectionPatch(partial=True)

DataTypeSchema.register_schema(collection_1_0_0_schema)
DataTypeSchema.register_schema(collection_1_1_0_schema)
DataTypeSchema.register_schema(collection_1_2_0_schema)
DataTypeSchema.register_schema(collection_1_3_0_schema)
DataTypeSchema.register_schema(collection_1_4_0_schema)
DataTypeSchema.register_schema(collection_1_5_0_schema)
DataTypeSchema.register_schema(collection_1_6_0_schema)
DataTypeSchema.register_schema(collection_1_7_0_schema)
DataTypeSchema.register_schema(collection_1_8_0_schema)
DataTypeSchema.register_schema(collection_1_9_0_schema)
