from typing import Dict
import uuid

import marshmallow as ma
from marshmallow_oneofschema import OneOfSchema

from .data_type import DataTypeField, DataTypeSchema, data_type_schema
from .identity_mapping import IdentityMapping
from ..patch_schema import create_patch_schema
from .path_restrictions import path_restrictions_schema
from .schema_version_checker import verify_schema_version, version_safe_type_schemas


class StorageGatewayPolicies(ma.fields.Method):
    def _jsonschema_type_mapping(self):
        type_mapping = {
            "oneOf": [],
        }
        refs = type_mapping["oneOf"]
        for policy in StorageGatewaySchema_1_0_0.connector_policies_schemas.values():
            title = getattr(policy, "title", policy.__class__.__name__)
            refs.append({"$ref": f"#/components/schemas/{title}"})

        metadata = self.metadata.get("metadata", {})
        metadata.update(self.metadata)
        for md_key, md_val in metadata.items():
            if md_key in ("metadata", "name"):
                continue
            type_mapping[md_key] = md_val
        return type_mapping


class StorageGatewaySchema_1_0_0(ma.Schema):
    __doc__ = """
    A storage gateway provides the access policies for the endpoint's
    connected storage systems. It is a named interface by which
    authorized users can create and manage collections on the
    connected storage system. A single storage system may be
    associated with multiple storage gateways, each with its own
    policies.

    Storage gateway policies describe what type connector the storage
    gateway uses, the paths it allows access to, the login
    requirements are for the storage gateway, and the algorithm to
    map Globus identities to the user namespace of the storage
    gateway (e.g. local accounts).
    """

    _DATA_TYPE = "storage_gateway#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    id = ma.fields.UUID(description="Unique id for this storage gateway")
    display_name = ma.fields.String(
        validate=[ma.validate.Length(max=256)],
        description="Name of the storage gateway",
    )
    connector_id = ma.fields.UUID(
        description="""
        Id of the connector type that this storage gateway interacts with.
        """
    )
    high_assurance = ma.fields.Boolean(
        required=False,
        allow_none=True,
        description="""
        Flag indicating if the storage_gateway requires high
        assurance features.
        """,
    )
    require_high_assurance = ma.fields.Boolean(
        required=False,
        allow_none=True,
        description="Alias for **high_assurance**",
        deprecated=True,
    )
    authentication_assurance_timeout = ma.fields.Int(
        required=False,
        allow_none=True,
        validate=[ma.validate.Range(min=1)],
        load_only=True,
        description="Alias for **authentication_timeout_mins**",
        deprecated=True,
    )
    authentication_timeout_mins = ma.fields.Int(
        required=False,
        allow_none=True,
        validate=[ma.validate.Range(min=1)],
        description="""
        Timeout (in minutes) during which a user is required to have
        authenticated to access files or create user credentials on this
        storage gateway.

        For a high assurance storage gateway, this must be done within the
        current Globus Auth session, otherwise, the caller can perform the
        authentication with any application which uses Globus Auth.
        """,
    )
    allowed_domains = ma.fields.List(
        ma.fields.String,
        description="""
        List of allowed domains. Users creating credentials or collections
        on this storage gateway must have an identity in one of these domains.
        """,
    )
    identity_mappings = ma.fields.List(
        ma.fields.Nested(IdentityMapping),
        required=False,
        allow_none=True,
        description="""
        List of identity mappings to apply to user identities to determine
        what connector-specific accounts are available for access.
        """,
    )
    users_allow = ma.fields.List(
        ma.fields.Str,
        required=False,
        allow_none=True,
        description="""
        List of connector-specific usernames allowed to access this storage
        gateway.
        """,
    )
    users_deny = ma.fields.List(
        ma.fields.Str,
        required=False,
        allow_none=True,
        description="""
        List of connector-specific usernames denied access to this storage
        gateway.
        """,
    )
    restrict_paths = ma.fields.Nested(
        path_restrictions_schema,
        required=False,
        allow_none=True,
        description="""
        Path restrictions within this storage gateway. Paths are
        interpreted as absolute paths in the file namespace of the
        connector.
        """,
    )
    policies = StorageGatewayPolicies(
        serialize="dump_policies",
        deserialize="load_policies",
        description="Connector-specific storage policies",
    )
    process_user = ma.fields.Str(
        required=False,
        allow_none=True,
        description="""
        Local POSIX user the GridFTP server should run as when accessing
        this storage gateway.
        """,
    )
    load_dsi_module = ma.fields.Str(
        required=False,
        allow_none=True,
        description="""
        Name of the DSI module to load by the GridFTP server when
        accessing this storage gateway.
        """,
    )
    deleted = ma.fields.Boolean(
        required=False,
        description="Flag indicating that this storage gateway has been deleted",
    )

    def load_policies(self, value):
        return data_type_schema.load(value, partial=self.partial)

    def dump_policies(self, obj):
        connector_id = str(ma.utils.get_value(obj, "connector_id"))
        if not self.load_only:
            schema = self.connector_policies_private_schemas.get(connector_id)
        else:
            schema = self.connector_policies_schemas.get(connector_id)
        if schema:
            return schema.dump(obj)
        else:
            raise ValueError("Unable to serialize {} {}".format(obj, connector_id))

    @ma.validates("DATA_TYPE")
    def validate_data_type(self, data, **kwargs):
        verify_schema_version(data, self._DATA_TYPE)

    @ma.post_load
    def backward_compatible_assurance_timeout_load(self, obj, **kwargs):
        authentication_assurance_timeout = obj.pop(
            "authentication_assurance_timeout", None
        )
        if authentication_assurance_timeout:
            obj["authentication_timeout_mins"] = authentication_assurance_timeout
        return obj

    @ma.post_dump
    def backward_compatible_require_high_assurance_dump(self, data, **kwargs):
        data["require_high_assurance"] = data.get("high_assurance")
        return data

    @ma.post_load
    def backward_compatible_require_high_assurance_load(self, obj, **kwargs):
        if "high_assurance" not in obj and "require_high_assurance" in obj:
            obj["high_assurance"] = obj.pop("require_high_assurance")
        return obj

    @ma.post_load
    def stringify_uuid(self, obj, **kwargs):
        res = {}
        for k, v in obj.items():
            if isinstance(v, uuid.UUID):
                res[k] = str(v)
            elif isinstance(v, list):
                newlist = []
                for i in v:
                    if isinstance(i, uuid.UUID):
                        newlist.append(str(i))
                    else:
                        newlist.append(i)
                res[k] = newlist
            else:
                res[k] = v
        return res

    # connectors may register their schema types in this dict, with
    # the connector ID as the key
    connector_policies_schemas = {}  # type: Dict[str, ma.Schema]
    connector_policies_private_schemas = {}  # type: Dict[str, ma.Schema]
    private = [
        "deleted",
        "identity_mappings",
        "restrict_paths",
        "process_user",
        "load_dsi_module",
        "users_allow",
        "users_deny",
        "authentication_assurance_timeout",  # private since it is deprecated
    ]

    @classmethod
    def register_schema(cls, public=None, private=None):
        if public:
            connector_id = public.connector_id
            cls.connector_policies_schemas[connector_id] = public
        if private:
            connector_id = private.connector_id
            cls.connector_policies_private_schemas[connector_id] = private


class StorageGatewaySchema_1_1_0(StorageGatewaySchema_1_0_0):
    __doc__ = (
        StorageGatewaySchema_1_0_0.__doc__
        + """
    Version 1.1.0 includes support for multi-factor authentication requirements
    for high assurance storage gateways.
    """
    )

    _DATA_TYPE = "storage_gateway#1.1.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    require_mfa = ma.fields.Boolean(
        required=False,
        default=False,
        description="""
        Flag indicating if the storage_gateway requires multi-factor
        authentication. Only usable on high assurance storage gateways.
        """,
    )


class StorageGatewaySchema_1_2_0(StorageGatewaySchema_1_1_0):
    __doc__ = (
        StorageGatewaySchema_1_1_0.__doc__
        + """
    Version 1.2.0 includes support for admin managed credentials.
    """
    )
    _DATA_TYPE = "storage_gateway#1.2.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    admin_managed_credentials = ma.fields.Boolean(
        required=False,
        default=False,
        description="""
        Flag indicating if the storage_gateway allows endpoint
        administrators to manage user credentials on behalf of other users.
        """,
    )


storage_gateway_1_0_0_schema = StorageGatewaySchema_1_0_0(
    load_only=StorageGatewaySchema_1_0_0.private
)
storage_gateway_1_0_0_private_schema = StorageGatewaySchema_1_0_0()
storage_gateway_1_0_0_patch_schema = create_patch_schema(StorageGatewaySchema_1_0_0)

storage_gateway_1_1_0_schema = StorageGatewaySchema_1_1_0(
    load_only=StorageGatewaySchema_1_1_0.private
)
storage_gateway_1_1_0_private_schema = StorageGatewaySchema_1_1_0()
storage_gateway_1_1_0_patch_schema = create_patch_schema(StorageGatewaySchema_1_1_0)

storage_gateway_1_2_0_schema = StorageGatewaySchema_1_2_0(
    load_only=StorageGatewaySchema_1_2_0.private
)
storage_gateway_1_2_0_private_schema = StorageGatewaySchema_1_2_0()
storage_gateway_1_2_0_patch_schema = create_patch_schema(StorageGatewaySchema_1_2_0)


class StorageGateway(OneOfSchema):
    __doc__ = StorageGatewaySchema_1_2_0.__doc__

    _DATA_TYPE = storage_gateway_1_2_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = version_safe_type_schemas(
        {
            "storage_gateway#1.0.0": storage_gateway_1_0_0_schema,
            "storage_gateway#1.1.0": storage_gateway_1_1_0_schema,
            "storage_gateway#1.2.0": storage_gateway_1_2_0_schema,
        }
    )
    type_field_remove = False

    def get_obj_type(self, obj):
        return self._DATA_TYPE

    @classmethod
    def register_schema(cls, public=None, private=None):
        for cls in cls.type_schemas.values():
            cls.register_schema(public=public, private=private)


class StorageGatewayPrivate(OneOfSchema):
    _DATA_TYPE = storage_gateway_1_2_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = {
        "storage_gateway#1.0.0": storage_gateway_1_0_0_private_schema,
        "storage_gateway#1.1.0": storage_gateway_1_1_0_private_schema,
        "storage_gateway#1.2.0": storage_gateway_1_2_0_private_schema,
    }
    type_field_remove = False

    def get_obj_type(self, obj):
        return self._DATA_TYPE


class StorageGatewayPatch(OneOfSchema):
    _DATA_TYPE = storage_gateway_1_2_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = version_safe_type_schemas(
        {
            "storage_gateway#1.0.0": storage_gateway_1_0_0_patch_schema,
            "storage_gateway#1.1.0": storage_gateway_1_1_0_patch_schema,
            "storage_gateway#1.2.0": storage_gateway_1_2_0_patch_schema,
        }
    )
    type_field_remove = False

    def get_obj_type(self, obj):
        return self._DATA_TYPE


storage_gateway_schema = StorageGateway(load_only=StorageGatewaySchema_1_0_0.private)
StorageGatewaySchema = StorageGateway
storage_gateway_private_schema = StorageGatewayPrivate()
storage_gateway_patch_schema = create_patch_schema(StorageGatewayPatch)

DataTypeSchema.register_schema(storage_gateway_1_0_0_schema)
DataTypeSchema.register_schema(storage_gateway_1_1_0_schema)
DataTypeSchema.register_schema(storage_gateway_1_2_0_schema)
