from typing import Dict

import marshmallow as ma

from .data_type import DataTypeField, DataTypeSchema, data_type_schema
from .schema_version_checker import verify_schema_version
from ..patch_schema import create_patch_schema


class CollectionPolicies(ma.fields.Method):
    def _jsonschema_type_mapping(self):
        type_mapping = {
            "oneOf": [],
        }
        refs = type_mapping["oneOf"]
        for policy in UserCredential.connector_policies_schemas.values():
            title = getattr(policy, "title", policy.__class__.__name__)
            refs.append({"$ref": f"#/components/schemas/{title}"})

        metadata = self.metadata.get("metadata", {})
        metadata.update(self.metadata)
        for md_key, md_val in metadata.items():
            if md_key in ("metadata", "name"):
                continue
            type_mapping[md_key] = md_val
        return type_mapping


class UserCredential(ma.Schema):
    _DATA_TYPE = "user_credential#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    id = ma.fields.UUID(
        required=False, description="Unique id for this user credential"
    )
    identity_id = ma.fields.UUID(
        description="Globus Auth identity id that this credential is associated with"
    )
    connector_id = ma.fields.UUID(
        description="Id of the connector type used by this credential"
    )
    storage_gateway_id = ma.fields.UUID(
        description="Storage Gateway this credential is associated with"
    )
    username = ma.fields.Str(
        description="""
        Connector-specific username that this credential is associated
        with. If the connector supports identity mapping, this matches the
        result of the mapping applied to identity_id.
        """
    )
    display_name = ma.fields.Str(
        required=False,
        allow_none=True,
        validate=[ma.validate.Length(max=256)],
        description="Display name of the credential",
    )
    home_directory = ma.fields.String(
        dump_only=True,
        description="The home directory of this account associated with this credential",
    )
    invalid = ma.fields.Boolean(
        required=False,
        description="Flag indicating that this credential is no longer valid",
    )
    provisioned = ma.fields.Boolean(
        required=False,
        description="""
        Flag indicating that this credential has been fully provisioned. If
        this is false and the invalid property is true, then the credential
        was created during login and patching it to add the missing data
        should be presented to the user as initializing the credential.
        """,
    )
    deleted = ma.fields.Boolean(
        required=False,
        description="Flag indicating that this credential has been deleted",
    )
    policies = CollectionPolicies(
        serialize="dump_policies",
        deserialize="load_policies",
        description="Connector-specific user credential policies",
    )

    private = ["home_directory"]

    connector_policies_schemas = {}  # type: Dict[str, ma.Schema]
    connector_policies_private_schemas = {}  # type: Dict[str, ma.Schema]

    @ma.validates("DATA_TYPE")
    def validate_data_type(self, data, **kwargs):
        verify_schema_version(data, self._DATA_TYPE)

    @ma.pre_load
    def ignore_loading_home_directory(self, data, **kwargs):
        """
        Remove the home_directory property when loading. We want to serialize
        this when replying to gridftp login but don't want to care about a
        remotely loaded value.
        """
        data.pop("home_directory", None)
        return data

    def load_policies(self, value):
        return data_type_schema.load(value, partial=self.partial, unknown=ma.EXCLUDE)

    def dump_policies(self, obj):
        connector_id = ma.utils.get_value(obj, "connector_id")

        if not self.load_only:
            schema = self.connector_policies_private_schemas.get(connector_id)
        else:
            schema = self.connector_policies_schemas.get(connector_id)
        if schema:
            return schema.dump(obj)
        else:
            raise ValueError("Unable to serialize {} {}".format(obj, connector_id))

    @ma.post_dump
    def remove_deleted(self, data, **kwargs):
        if not data.get("deleted"):
            data.pop("deleted", None)
        return data

    @classmethod
    def register_schema(cls, public=None, private=None):
        if public:
            connector_id = public.connector_id
            cls.connector_policies_schemas[connector_id] = public
            if not private:
                cls.connector_policies_private_schemas[connector_id] = public
        if private:
            connector_id = private.connector_id
            cls.connector_policies_private_schemas[connector_id] = private


user_credential_schema = UserCredential(load_only=UserCredential.private)
UserCredentialSchema = UserCredential
user_credential_private_schema = UserCredential()
user_credential_patch_schema = create_patch_schema(UserCredential)
DataTypeSchema.register_schema(user_credential_schema)
