from typing import List

import marshmallow as ma

from ..api.collection import CollectionSchema
from ..api.data_type import DataTypeField, DataTypeSchema
from ..api.storage_gateway import StorageGatewaySchema
from ..api.user_credential import UserCredentialSchema


HPSS_CONNECTOR_ID = "fb656a17-0f69-4e59-95ff-d0a62ca7bdf5"


class HPSSStoragePolicies(ma.Schema):
    """
    Connector-specific storage gateway policies for the HPSS connector
    """

    _DATA_TYPE = "hpss_storage_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    authentication_mech = ma.fields.String(
        validate=[ma.validate.OneOf(["krb5", "unix"])],
        required=True,
        description="""
        The type of authentication the connector will perform when logging into HPSS
        """,
    )
    authenticator = ma.fields.Str(
        required=True,
        validate=ma.validate.Regexp(r"^(auth_keyfile|auth_keytab):/"),
        description="""
        Authenticator used with authentication mech to perform authentication
        to HPSS. Format is: "<auth_type>:<auth_file>" where <auth_type> is one
        of "auth_keytab" or "auth_keyfile".
        """,
        example="auth_keytab:/var/hpss/etc/gridftp.keytab",
    )
    uda_checksum = ma.fields.Boolean(
        required=True,
        description="""
        Flag that indicates if checksums should be stored within UDAs so that
        sync-by-checksum transfers can verify the file without staging the file
        from tape.
        """,
    )

    private = ["authentication_mech", "authenticator"]
    connector_id = HPSS_CONNECTOR_ID


class HPSSCollectionPolicies(ma.Schema):
    """
    Connector-specific collection policies for the HPSS connector
    """

    _DATA_TYPE = "hpss_collection_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    private: List[str] = []
    connector_id = HPSS_CONNECTOR_ID


class HPSSUserCredentialPolicies(ma.Schema):
    """
    Connector-specific user credential policies for the HPSS connector
    """

    _DATA_TYPE = "hpss_user_credential_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    private: List[str] = []
    connector_id = HPSS_CONNECTOR_ID


hpss_storage_policies_private_schema = HPSSStoragePolicies()
hpss_storage_policies_schema = HPSSStoragePolicies(
    load_only=HPSSStoragePolicies.private
)
DataTypeSchema.register_schema(hpss_storage_policies_schema)
StorageGatewaySchema.register_schema(
    public=hpss_storage_policies_schema,
    private=hpss_storage_policies_private_schema,
)
hpss_collection_policies_private_schema = HPSSCollectionPolicies()
hpss_collection_policies_schema = HPSSCollectionPolicies(
    load_only=HPSSCollectionPolicies.private
)
DataTypeSchema.register_schema(hpss_collection_policies_schema)
CollectionSchema.register_schema(
    public=hpss_collection_policies_schema,
    private=hpss_collection_policies_private_schema,
)
hpss_user_credential_policies_private_schema = HPSSUserCredentialPolicies()
hpss_user_credential_policies_schema = HPSSUserCredentialPolicies(
    load_only=HPSSUserCredentialPolicies.private
)
DataTypeSchema.register_schema(hpss_user_credential_policies_schema)
UserCredentialSchema.register_schema(
    public=hpss_user_credential_policies_schema,
    private=hpss_user_credential_policies_private_schema,
)
