import marshmallow as ma

from typing import List
from ..cleanup_none_mixin import CleanupNoneMixin
from ..api.collection import CollectionSchema
from ..api.data_type import DataTypeField, DataTypeSchema
from ..api.storage_gateway import StorageGatewaySchema
from ..api.user_credential import UserCredentialSchema


IRODS_CONNECTOR_ID = "e47b6920-ff57-11ea-8aaa-000c297ab3c2"


class IrodsEnvironment(ma.Schema):
    """
    Variables to set in the iRODS client environment.
    """

    name = ma.fields.String(required=True, description="Environment variable name")
    value = ma.fields.String(required=True, description="Environment variable value")


class IrodsStoragePolicies(ma.Schema, CleanupNoneMixin):
    """
    Connector-specific storage gateway policies for the Irods connector
    """

    _DATA_TYPE = "irods_storage_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)
    irods_environment_file = ma.fields.String(
        required=True,
        description="Path to the irods environment file",
        example="/var/irods/irods_environment.json",
    )
    irods_authentication_file = ma.fields.String(
        description="Path to the irods authentication file",
        example="/var/irods/.irodsA",
    )
    environment = ma.fields.List(
        ma.fields.Nested(IrodsEnvironment()),
        description="Variables to set in the iRODS client environment",
    )
    private = ["irods_environment_file", "irods_authentication_file", "environment"]
    connector_id = IRODS_CONNECTOR_ID


class IrodsCollectionPolicies(ma.Schema):
    """
    Connector-specific collection policies for the Irods connector
    """

    _DATA_TYPE = "irods_collection_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    connector_id = IRODS_CONNECTOR_ID
    private = []  # type: List[str]


class IrodsUserCredentialPolicies(ma.Schema):
    """
    Connector-specific user credential policies for the Irods connector
    """

    _DATA_TYPE = "irods_user_credential_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    connector_id = IRODS_CONNECTOR_ID
    private = []  # type: List[str]


irods_storage_policies_schema = IrodsStoragePolicies(
    load_only=IrodsStoragePolicies.private
)
irods_storage_policies_private_schema = IrodsStoragePolicies()
irods_user_credential_policies_schema = IrodsUserCredentialPolicies(
    load_only=IrodsUserCredentialPolicies.private
)
irods_collection_policies_schema = IrodsCollectionPolicies(
    load_only=IrodsCollectionPolicies.private
)
irods_collection_policies_private_schema = IrodsCollectionPolicies()

DataTypeSchema.register_schema(irods_storage_policies_schema)
DataTypeSchema.register_schema(irods_user_credential_policies_schema)
DataTypeSchema.register_schema(irods_collection_policies_schema)

StorageGatewaySchema.register_schema(
    public=irods_storage_policies_schema, private=irods_storage_policies_private_schema
)

UserCredentialSchema.register_schema(public=irods_user_credential_policies_schema)

CollectionSchema.register_schema(
    public=irods_collection_policies_schema,
    private=irods_collection_policies_private_schema,
)
