from typing import List

import marshmallow as ma
from marshmallow_oneofschema import OneOfSchema

from ..api.collection import CollectionSchema
from ..api.data_type import DataTypeField, DataTypeSchema
from ..api.storage_gateway import StorageGatewaySchema
from ..api.user_credential import UserCredentialSchema


S3_CONNECTOR_ID = "7643e831-5f6c-4b47-a07f-8ee90f401d23"


class S3StoragePolicies_1_0_0(ma.Schema):
    __doc__ = """
    Connector-specific storage gateway policies for the S3 connector
    """

    _DATA_TYPE = "s3_storage_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    s3_endpoint = ma.fields.Url(
        description="URL of the S3 API endpoint",
        example="https://s3.amazonaws.com",
    )
    s3_buckets = ma.fields.List(
        ma.fields.String,
        description="""
        List of buckets not owned by the collection owner that will be shown in
        the root of collections created at the base of this storage gateway.
        """,
    )
    s3_user_credential_required = ma.fields.Boolean(
        description="""
        Flag indicating if a Globus User must register a user credential in
        order to create a guest collection on this storage gateway.
        """
    )
    private: List[str] = []
    connector_id = S3_CONNECTOR_ID


class S3StoragePolicies_1_1_0(S3StoragePolicies_1_0_0):
    __doc__ = (
        S3StoragePolicies_1_0_0.__doc__
        + """
    Version 1.1.0 adds support for the s3_requester_pays property
    """
    )

    _DATA_TYPE = "s3_storage_policies#1.1.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    s3_requester_pays = ma.fields.Boolean(
        description="""
        Flag indicating that S3 operations will be charged to the account of
        the registered credentials. Credentials used with a storage gateway
        that has the s3_requester_pays property set to true are invalid unless
        they also have this property set to true as an acknowledgement.
        """
    )
    private: List[str] = []
    connector_id = S3_CONNECTOR_ID


s3_storage_policies_1_0_0_schema = S3StoragePolicies_1_0_0(
    load_only=S3StoragePolicies_1_0_0.private
)
s3_storage_policies_1_1_0_schema = S3StoragePolicies_1_1_0(
    load_only=S3StoragePolicies_1_1_0.private
)
s3_storage_policies_1_0_0_private_schema = S3StoragePolicies_1_0_0()
s3_storage_policies_1_1_0_private_schema = S3StoragePolicies_1_1_0()


class S3StoragePolicies(OneOfSchema):
    __doc__ = s3_storage_policies_1_1_0_schema.__doc__

    _DATA_TYPE = s3_storage_policies_1_1_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = {
        "s3_storage_policies#1.0.0": s3_storage_policies_1_0_0_schema,
        "s3_storage_policies#1.1.0": s3_storage_policies_1_1_0_schema,
    }
    type_field_remove = False
    connector_id = S3_CONNECTOR_ID
    title = "S3StoragePolicies"

    def get_obj_type(self, obj):
        return self._DATA_TYPE

    private = s3_storage_policies_1_1_0_schema.private + []


class S3StoragePoliciesPrivate(OneOfSchema):
    __doc__ = s3_storage_policies_1_1_0_schema.__doc__
    _DATA_TYPE = s3_storage_policies_1_1_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = {
        "s3_storage_policies#1.0.0": s3_storage_policies_1_0_0_private_schema,
        "s3_storage_policies#1.1.0": s3_storage_policies_1_1_0_private_schema,
    }
    type_field_remove = False
    connector_id = S3_CONNECTOR_ID
    title = "S3StoragePolicies"

    def get_obj_type(self, obj):
        return self._DATA_TYPE


class S3CollectionPolicies(ma.Schema):
    __doc__ = s3_storage_policies_1_1_0_schema.__doc__

    _DATA_TYPE = "s3_collection_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    private: List[str] = []
    connector_id = S3_CONNECTOR_ID


s3_collection_policies_schema = S3CollectionPolicies(
    load_only=S3CollectionPolicies.private
)
s3_collection_policies_private_schema = S3CollectionPolicies()


class S3UserCredentialPolicies_1_0_0(ma.Schema):
    __doc__ = """
    Connector-specific user credential policies for the S3 connector
    """

    _DATA_TYPE = "s3_user_credential_policies#1.0.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    s3_key_id = ma.fields.String(
        allow_none=True,
        description="""
        Access Key ID to use with the S3 API to access your buckets and
        objects.
        """,
    )
    s3_secret_key = ma.fields.String(
        allow_none=True,
        description="""
        Secret key to use with the S3 API to access your buckets and objects.
        """,
    )

    private = ["s3_secret_key"]
    connector_id = S3_CONNECTOR_ID


class S3UserCredentialPolicies_1_1_0(S3UserCredentialPolicies_1_0_0):
    __doc__ = (
        S3UserCredentialPolicies_1_0_0.__doc__
        + """
    Version 1.1.0 adds support for the s3_requester_pays property.
    """
    )

    _DATA_TYPE = "s3_user_credential_policies#1.1.0"
    DATA_TYPE = DataTypeField(_DATA_TYPE)

    s3_requester_pays = ma.fields.Boolean(
        description="""
        Flag indicating the user acknowledges S3 operations will be charged to
        the account of this credential.  If this flag is true in the storage
        gateway policy, this must also be true or the credential will be invalid.
        """
    )


s3_user_credential_policies_1_0_0_schema = S3UserCredentialPolicies_1_0_0(
    load_only=S3UserCredentialPolicies_1_0_0.private
)
s3_user_credential_policies_1_1_0_schema = S3UserCredentialPolicies_1_1_0(
    load_only=S3UserCredentialPolicies_1_1_0.private
)
s3_user_credential_policies_1_0_0_private_schema = S3UserCredentialPolicies_1_0_0()
s3_user_credential_policies_1_1_0_private_schema = S3UserCredentialPolicies_1_1_0()


class S3UserCredentialPolicies(OneOfSchema):
    __doc__ = S3UserCredentialPolicies_1_1_0.__doc__

    _DATA_TYPE = s3_user_credential_policies_1_1_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = {
        "s3_user_credential_policies#1.0.0": s3_user_credential_policies_1_0_0_schema,
        "s3_user_credential_policies#1.1.0": s3_user_credential_policies_1_1_0_schema,
    }
    type_field_remove = False
    connector_id = S3_CONNECTOR_ID
    title = "S3UserCredentialPolicies"

    def get_obj_type(self, obj):
        return self._DATA_TYPE

    private = s3_user_credential_policies_1_1_0_schema.private + []


class S3UserCredentialPoliciesPrivate(OneOfSchema):
    __doc__ = S3UserCredentialPolicies_1_1_0.__doc__

    _DATA_TYPE = s3_user_credential_policies_1_1_0_schema._DATA_TYPE
    type_field = "DATA_TYPE"
    type_schemas = {
        "s3_user_credential_policies#1.0.0": s3_user_credential_policies_1_0_0_private_schema,
        "s3_user_credential_policies#1.1.0": s3_user_credential_policies_1_1_0_private_schema,
    }
    type_field_remove = False
    connector_id = S3_CONNECTOR_ID
    title = "S3UserCredentialPolicies"

    def get_obj_type(self, obj):
        return self._DATA_TYPE


s3_storage_policies_private_schema = S3StoragePoliciesPrivate()
s3_storage_policies_schema = S3StoragePolicies(load_only=S3StoragePolicies.private)

s3_user_credential_policies_private_schema = S3UserCredentialPoliciesPrivate()
s3_user_credential_policies_schema = S3UserCredentialPolicies(
    load_only=S3UserCredentialPolicies.private
)

s3_collection_policies_private_schema = S3CollectionPolicies()
s3_collection_policies_schema = S3CollectionPolicies(
    load_only=S3CollectionPolicies.private
)


DataTypeSchema.register_schema(s3_storage_policies_1_0_0_schema)
DataTypeSchema.register_schema(s3_storage_policies_1_1_0_schema)
DataTypeSchema.register_schema(s3_user_credential_policies_1_0_0_schema)
DataTypeSchema.register_schema(s3_user_credential_policies_1_1_0_schema)
DataTypeSchema.register_schema(s3_collection_policies_schema)

StorageGatewaySchema.register_schema(
    public=s3_storage_policies_schema, private=s3_storage_policies_private_schema
)

UserCredentialSchema.register_schema(
    public=s3_user_credential_policies_schema,
    private=s3_user_credential_policies_private_schema,
)

CollectionSchema.register_schema(
    public=s3_collection_policies_schema, private=s3_collection_policies_private_schema
)
