"""
globus-connect-server endpoint setup
"""
import json
import os
import datetime
import textwrap
import typing as t
import uuid

import click
from flask import current_app
import globus_sdk
from globus_sdk.scopes import AuthScopes, TransferScopes
import jwcrypto

from globus.manager.constants import RoleType
from globus.manager.api.models import (
    db,
    AcmeAccount,
    LetsEncryptCertificate,
    GlobusDnsDomain,
    Endpoint,
    Role,
)
from globus.manager.appsync.query import GraphQLError
import globus.manager.api.errors as errors
from globus.manager.appsync.client import ConfigurationObjectTag
from globus.manager.utils.keychain import (
    DataKeyError,
    NodeKey,
)
from globus.manager.version import __version_fingerprint__

from globus.manager.api.app import make_app
from gcs_cli.parsing.types import EXPLICIT_NULL
from gcs_cli.config import internal_auth_client
from gcs_cli import version as gcs_cli_version


def endpoint_setup(
    display_name: str,
    client_id: t.Optional[str],
    secret: t.Optional[str],
    deployment_key_path: str,
    deployment_key: t.Optional[NodeKey],
    project_id: t.Optional[str],
    owner: str,
    agree_to_letsencrypt_tos: bool,
    dont_set_advertised_owner: bool,
    contact_email: str,
    development_appsync: t.Optional[t.Tuple[str, str]] = None,
    **kwargs,
) -> None:

    ctx = click.get_current_context()

    if not agree_to_letsencrypt_tos:
        click.echo(
            "\n".join(
                textwrap.wrap(
                    "Globus Connect Server uses the Let's Encrypt service to "
                    "manage certificates for the web services it installs "
                    "on your data transfer nodes. These certificates are "
                    "issued for DNS domain names operated by the Globus "
                    "Project."
                )
                + [""]
                + textwrap.wrap("Please read the Terms of Service at:")
                + ["https://letsencrypt.org/repository/"]
            )
        )

        if not click.confirm("Do you agree to these Terms of Service?"):
            ctx.exit(1)

    app = make_app(app_config="globus.manager.api.config.endpoint_cli")
    db.init_app(app)

    endpoint: t.Dict[str, t.Any] = {
        "DATA_TYPE": "endpoint#1.0.0",
        "contact_email": contact_email,
    }
    for k, v in kwargs.items():
        if v is None:
            continue
        if v == EXPLICIT_NULL:
            endpoint[k] = None
        elif isinstance(v, uuid.UUID):
            endpoint[k] = str(v)
        else:
            endpoint[k] = v

    endpoint["display_name"] = display_name
    environment = os.environ.get("GLOBUS_SDK_ENVIRONMENT", "production")

    with app.app_context():
        db.create_all()
        db.session.commit()

        app.config["CLIENT_ID"] = str(client_id)
        app.config["NODE_ID"] = str(client_id)
        app.config["CLIENT_SECRET"] = secret
        app.config["ENVIRONMENT"] = environment
        if development_appsync:
            app.config["GCS_APPSYNC_URL"] = development_appsync[0]
            app.config["GCS_APPSYNC_CLIENT_ID"] = development_appsync[1]

        endpoint_setup = EndpointSetup(
            ctx,
            client_id,
            secret,
            deployment_key_path,
            deployment_key,
            project_id,
            owner,
            endpoint,
            environment,
            contact_email,
            dont_set_advertised_owner,
        )
        endpoint_setup.run()
        click.echo(
            "Created endpoint {}".format(endpoint_setup.endpoint_model.endpoint_id)
        )

        click.echo("Endpoint domain_name {}".format(endpoint_setup.domain))

        if not endpoint_setup.subscription_set:
            click.echo(
                "No subscription is set on this endpoint, so only basic "
                "features are enabled.\n"
            )

            click.echo(
                "\nTo enable subscription features on this endpoint, have "
                "the\nGlobus subscription manager for your organization "
                "assign a subscription to this endpoint.\n"
            )

        click.echo(
            "\nIf you plan on using the Google Drive or Google Cloud Storage\n"
            "connectors, use\n "
            "    https://{}/api/v1/authcallback_google\n"
            "as the Authorized redirect URI for this endpoint".format(
                endpoint_setup.domain
            )
        )


class EndpointSetup:
    def __init__(
        self,
        ctx,
        client_id: t.Optional[str],
        secret: t.Optional[str],
        deployment_key_path,
        deployment_key,
        project_id: t.Optional[str],
        owner,
        endpoint,
        environment,
        contact_email: t.Optional[str],  # TODO: missing when called from migrate53
        dont_set_advertised_owner: bool,
    ) -> None:
        self.ctx = ctx
        self.client_id = client_id
        self.secret = secret
        self.deployment_key_path = deployment_key_path
        self.deployment_key = deployment_key
        self.project_id = project_id
        self.endpoint = endpoint
        self.subscription_id = endpoint.pop("subscription_id", None)
        self.subscription_set = False
        self.environment = environment
        self.domain = None
        self.owner = owner
        self.contact_email = contact_email
        self.dont_set_advertised_owner = dont_set_advertised_owner

    def run(self):
        tasks = [
            # START: Bootstrap phase, possibly no client ID or secret
            self.verify_owner_id,
            self.do_link_auth_flow,
            self.verify_client_credentials,
            self.get_client_credentials,
            # END: Bootstrap phase, possibly no client ID or secret
            self.load_or_create_deployment_key,
            self.check_for_existing_keychain,
            self.create_initial_endpoint,
            self.add_keys_to_keychain,
            self.download_existing_configuration,
            self.create_transfer_endpoint,
            self.check_transfer_subscription,
            self.create_roles,
            self.pre_register_dns,
            self.register_dns,
            self.create_acme_client,
            self.pre_provision_certificate,
            self.provision_certificate,
            self.register_server,
        ]

        if not self.dont_set_advertised_owner:
            tasks.append(self.set_owner)

        def show(obj):
            if obj and obj.__doc__:
                return obj.__doc__.strip().split("\n")[0]
            return ""

        with click.progressbar(tasks, item_show_func=show, width=25) as bar:
            for task in bar:
                task()

    # START: Bootstrap phase, possibly no client ID or secret

    def verify_owner_id(self):
        """
        Verify owner ID.

        Verifies self.owner and sets self.owner_id (str)
        """
        current_app.logger.info("verify_owner_id: start")
        auth_client = internal_auth_client()
        response = auth_client.get_identities(usernames=[self.owner])
        owner_identities = response.data.get("identities", [])
        assert len(owner_identities) == 0 or len(owner_identities) == 1

        if len(owner_identities) != 1:
            click.echo(
                "\nUnable to verify that Globus user {} exists".format(self.owner)
            )
            self.ctx.exit(1)

        owner_identity = owner_identities[0]
        current_app.logger.debug(
            "verify_owner_id: username=%s id=%s status=%s",
            self.owner,
            owner_identity["id"],
            owner_identity["status"],
        )
        if owner_identity["status"] not in ["used", "private"]:
            click.echo(
                "\nRequested owner {} exists but does not have a "
                "validated account".format(self.owner)
            )
            self.ctx.exit(1)

        if owner_identity["id"] == str(self.client_id):
            click.echo(
                "\nThe parameter to the --owner command-line option must be "
                "different than the endpoint's client id",
                err=True,
            )
            self.ctx.exit(1)

        if owner_identity["identity_type"] == "email":
            click.echo(
                "\n"
                + "\n".join(
                    textwrap.wrap(
                        "The parameter to the --owner command-line option "
                        "must be an account with a non-email identity_type",
                    )
                ),
                err=True,
            )
            self.ctx.exit(1)

        self.owner_id = owner_identity["id"]
        current_app.logger.info("verify_owner_id: end")

    def do_link_auth_flow(self) -> None:
        """
        Perform any necessary Auth flows.

        Starts interactive flow for tokens to be used with set_owner() and
        get_client_credentials(). Outputs self.transfer_token and self.auth_token.

        TODO: We use internal_auth_client which uses the CLI's template client which
        acts on its own behalf (not the callers). We should try to reuse tokens that
        are already in token storage.
        """
        current_app.logger.info("do_auth_link_flow: start")

        scopes = []
        session_params = {"prompt": "login"}
        session_messages = []

        if not self.client_id:
            scopes += ["urn:globus:auth:scope:auth.globus.org:manage_projects"]
            session_messages.append("register the new endpoint to get credentials ")

        if self.dont_set_advertised_owner is False:
            scopes += [TransferScopes.all]
            session_messages.append("set the advertised owner of the endpoint")
            session_params["session_required_identities"] = self.owner_id

        if scopes:
            auth_client = internal_auth_client()
            try:
                auth_client.oauth2_start_flow(
                    redirect_uri=auth_client.base_url + "v2/web/auth-code",
                    refresh_tokens=False,
                    requested_scopes=scopes,
                )
                session_message = "and ".join(session_messages)
                session_params["session_message"] = session_message.capitalize()
                linkprompt = f"Please authenticate with Globus to {session_message}"
                click.echo(
                    "\n{0}:\n{1}\n{2}\n{1}\n".format(
                        linkprompt,
                        "-" * len(linkprompt),
                        auth_client.oauth2_get_authorize_url(
                            query_params=session_params
                        ),
                    )
                )
                auth_code = click.prompt(
                    "Enter the resulting Authorization Code here", default=""
                ).strip()
                if not auth_code:
                    raise Exception("No auth code")
                token_response = auth_client.oauth2_exchange_code_for_tokens(auth_code)
            except Exception as e:
                raise click.ClickException(f"Authorization flow failed: {e}")

            self.transfer_token = token_response.by_resource_server.get(
                TransferScopes.resource_server
            )
            self.auth_token = token_response.by_resource_server.get(
                AuthScopes.resource_server
            )

        current_app.logger.info("do_auth_link_flow: end")

    def verify_client_credentials(self) -> None:
        """
        Verify client credentials
        """
        if self.client_id:
            try:
                appsync_client = current_app.appsync_client
                if appsync_client is None:
                    click.echo("Unable to contact Globus services", err=True)
                    self.ctx.exit(1)
            except globus_sdk.GlobusAPIError as e:
                current_app.logger.error(e)
                click.echo(
                    "\n".join(
                        textwrap.wrap(
                            "Unable to contact Globus services with provided client "
                            "id and secret. Please verify that those values are "
                            "correct."
                        )
                    ),
                    err=True,
                )
                self.ctx.exit(1)

    def get_client_credentials(self) -> None:
        """
        Register the endpoint with Globus Auth

        Outputs:
        - self.client_id
        - self.secret
        - app.config["CLIENT_ID"] = str(client_id)
        - app.config["NODE_ID"] = str(client_id)
        - app.config["CLIENT_SECRET"] = secret
        """
        current_app.logger.info("get_client_credentials: start")

        if not self.client_id:
            assert self.auth_token

            client = globus_sdk.AuthClient(
                authorizer=globus_sdk.AccessTokenAuthorizer(
                    self.auth_token["access_token"]
                ),
                app_name=gcs_cli_version.app_name,
            )

            if not self.project_id:
                try:
                    projects_response = client.get("/v2/api/projects")
                except globus_sdk.GlobusAPIError as e:
                    current_app.logger.warning(e)
                    raise click.ClickException(f"Failed to get the project list: {e}")

                if len(projects_response["projects"]) > 1:
                    raise click.ClickException(
                        "You are the admin of multiple projects, please use the "
                        "--project-id option to specify which project to use. "
                        "Current projects:\n"
                        + "\n".join(
                            [
                                f"{d['id']} : {d['project_name']}"
                                for d in projects_response["projects"]
                            ]
                        )
                    )

                elif len(projects_response["projects"]) == 1:
                    self.project_id = projects_response["projects"][0]["id"]

                else:
                    # Create a default project
                    project_doc = {
                        "project": {
                            "display_name": "Globus Connect Server Endpoints",
                            "contact_email": self.contact_email,
                            "admin_ids": [str(self.owner_id)],
                        }
                    }
                    try:
                        response = client.post("/v2/api/projects", data=project_doc)
                    except globus_sdk.GlobusAPIError as e:
                        current_app.logger.warning(e)
                        raise click.ClickException(
                            f"Failed to get create a project: {e}"
                        )

                    self.project_id = response["project"]["id"]

            # Create the client
            client_doc = {
                "client": {
                    "name": self.endpoint["display_name"],
                    "project": str(self.project_id),
                    "public_client": False,
                    "client_type": "globus_connect_server",
                }
            }

            try:
                response = client.post("/v2/api/clients", data=client_doc)
            except globus_sdk.GlobusAPIError as e:
                current_app.logger.warning(e)
                raise click.ClickException(f"Failed to get create a client: {e}")
            client_id = response["client"]["id"]

            # Create the credentials
            credential_doc = {
                "credential": {
                    "name": f"Created by {gcs_cli_version.app_name} at {datetime.datetime.now()}",
                }
            }

            try:
                response = client.post(
                    f"/v2/api/clients/{client_id}/credentials", data=credential_doc
                )
            except globus_sdk.GlobusAPIError as e:
                current_app.logger.warning(e)
                raise click.ClickException(
                    f"Failed to get create client credentials: {e}"
                )
            secret = response["credential"]["secret"]

            self.client_id = str(client_id)
            self.secret = secret
            current_app.config["CLIENT_ID"] = str(client_id)
            current_app.config["NODE_ID"] = str(client_id)
            current_app.config["CLIENT_SECRET"] = secret

        current_app.logger.info("get_client_credentials: end")

    # END: Bootstrap phase, possibly no client ID or secret

    def load_or_create_deployment_key(self):
        """
        Configure deployment key

        If the output deployment key exists, load it and use that as the
        deployment key. Otherwise create a new key and serialize it to that
        file.

        TODO: 'load' is now performed via key_and_credentials_options but other
        code (ex. migrate53) still require this path. And all paths need to have
        self.node_key and current_app.config["DEPLOYMENT_KEY"].
        """
        if not self.deployment_key:
            # TODO: Remove thie block once migrate53 is removed
            if os.path.exists(self.deployment_key_path):
                current_app.logger.debug("Trying to read existing node key")
                # If deployment key is already persisted, use it
                with open(self.deployment_key_path, "r") as f:
                    try:
                        self.deployment_key = NodeKey(**json.load(f))
                    except Exception as e:
                        raise DataKeyError(
                            f"Invalid key file {self.deployment_key_path}", e
                        )

            if self.deployment_key is None:
                # If deployment key is not persisted, create and persist it locally
                self.deployment_key = NodeKey.generate()
                old_umask = os.umask(0o77)
                with open(self.deployment_key_path, "w") as f:
                    data = {
                        **self.deployment_key.export(private_key=True, as_dict=True),
                        "client_id": self.client_id,
                        "secret": self.secret,
                    }
                    try:
                        json.dump(data, f)
                    except IOError as e:
                        raise DataKeyError(
                            f"Unable to write key file {self.deployment_key_path}", e
                        )
                    finally:
                        os.umask(old_umask)

        # TODO: Is self.node_key referenced from outside 'endpoint setup'? If no,
        # then remove it.
        self.node_key = self.deployment_key
        current_app.config["DEPLOYMENT_KEY"] = self.deployment_key

    def check_for_existing_keychain(self):
        """
        Configure keychain
        """
        current_app.logger.info("check_for_existing_keychain: start")
        try:
            items = current_app.appsync_client.query_all(
                id=ConfigurationObjectTag.KEY_CHAIN.value
            )
            for item in items:
                current_app.appsync_client.process_item(item, catch_exceptions=False)
        except GraphQLError as e:
            current_app.logger.info(str(e))
        except jwcrypto.jwe.InvalidJWEData as e:
            current_app.logger.error(str(e))
            click.echo(
                "\n"
                + "\n".join(
                    textwrap.wrap(
                        "Error decrypting keychain. This usually occurs when an "
                        "endpoint was previously set up with the same client ID "
                        "but a different deployment key. If you still have the "
                        "old deployment key, you can rerun this command with the "
                        "--deployment-key parameter to use it. If not, you should "
                        "run globus-connect-server endpoint cleanup, "
                        "create a new endpoint Auth client, and then rerun this "
                        "command."
                    )
                ),
                err=True,
            )
            self.ctx.exit(1)

        current_app.logger.info("check_for_existing_keychain: end")

    def create_initial_endpoint(self):
        """
        Initialize Endpoint
        """
        current_app.logger.info("create_initial_endpoint: start")
        try:
            current_app.logger.debug("create_initial_endpoint: query_appsync")
            items = current_app.appsync_client.query_all(
                tag=ConfigurationObjectTag.ENDPOINT
            )
            if items:
                endpoint_model = current_app.appsync_client.process_item(items[0])
            else:
                endpoint_model = None
            current_app.logger.debug(
                "create_initial_endpoint: query_result=%s", endpoint_model
            )
        except GraphQLError as e:
            current_app.logger.info(str(e))
            endpoint_model = None

        if not endpoint_model:
            current_app.logger.debug("create_initial_endpoint: create_stub_endpoint")
            endpoint_model = Endpoint(doc={})
            try:
                current_app.logger.debug(
                    "create_initial_endpoint: register_with_appsync"
                )
                current_app.appsync_client.create_object(endpoint_model)
            except GraphQLError as e:
                if e.data:
                    res = e.data
                    for o in res.get("configuration_objects", {}).get("items", []):
                        if (
                            ConfigurationObjectTag(o["tag"])
                            == ConfigurationObjectTag.ENDPOINT
                            and o["id"] == self.client_id
                        ):
                            endpoint_model = Endpoint(appsync_data=o)
        self.endpoint_model = endpoint_model
        current_app.logger.debug("create_initial_endpoint: add_to_database")
        db.session.add(endpoint_model)
        current_app.logger.debug("create_initial_endpoint: commit_to_database")
        db.session.commit()
        current_app.logger.info("create_initial_endpoint: end")

    def add_keys_to_keychain(self):
        """
        Add keys to keychain
        """
        current_app.logger.info("add_keys_to_keychain: start")
        if self.deployment_key.thumbprint() not in [
            k.thumbprint() for k in current_app.keychain.node_keys
        ]:
            current_app.logger.info(
                "add_keys_to_keychain: add key_id=%s", self.deployment_key.thumbprint()
            )
            try:
                current_app.keychain.add_node_key(self.deployment_key)
            except errors.LimitExceededError as e:
                click.echo(f"Error adding node key: {e}", err=True)
                self.ctx.exit(1)

        if not current_app.keychain.data_keys:
            dk = current_app.keychain.add_data_key()
            current_app.logger.info(
                "add_keys_to_keychain: create_data_key key_id=%s", dk.thumbprint()
            )
        try:
            current_app.appsync_client.create_object(current_app.keychain)
        except GraphQLError as e:
            conflict_error = False
            for errobj in e.errors:
                if errobj["errorType"] == "CONFLICT":
                    conflict_error = True
            if not conflict_error:
                current_app.logger.error(str(e))
                self.ctx.exit(1)
        current_app.logger.info("add_keys_to_keychain: end")

    @staticmethod
    def __endpoint_config_filter(items):
        return [
            i
            for i in items
            if (
                i["tag"] == ConfigurationObjectTag.ENDPOINT
                or i["tag"] == ConfigurationObjectTag.WILDCARD_DOMAIN
                or i["tag"] == ConfigurationObjectTag.ACME_ACCOUNT
                or i["tag"] == ConfigurationObjectTag.CERTIFICATE
                or (i["tag"] == ConfigurationObjectTag.ROLE and not i.get("resource"))
            )
        ]

    def download_existing_configuration(self):
        """
        Download configuration data
        """

        current_app.logger.info("download_existing_configuration: start")
        # Download endpoint configuration if present
        current_app.logger.info("download_existing_configuration: query_models")
        items = current_app.appsync_client.query_all(
            item_filter=self.__endpoint_config_filter
        )
        current_app.logger.info(
            "download_existing_configuration: commit_models_to_session"
        )
        current_app.appsync_client.commit_items_to_session(db.session, items)
        current_app.logger.info("download_existing_configuration: end")

    def create_transfer_endpoint(self):
        """
        Create Endpoint in Globus
        """
        current_app.logger.info("create_transfer_endpoint: start")
        endpoint_model = Endpoint.query.one()

        if not hasattr(endpoint_model, "endpoint_id"):
            try:
                endpoint_id = self.endpoint.get("id", self.client_id)
                res = current_app.transfer_client.get_endpoint(endpoint_id)
                if res:
                    current_app.logger.debug(
                        "create_transfer_endpoint: endpoint already exists, "
                        "but is not in gcs.globus.org"
                    )
                    endpoint_model.endpoint_id = endpoint_id
            except globus_sdk.TransferAPIError as e:
                if e.code == "EndpointDeleted":
                    click.echo(
                        "\nAn endpoint with this ID has already been deleted "
                        "and this ID may not be reused.\nRegister a new "
                        "GCS Endpoint to create the endpoint",
                        err=True,
                    )
                    self.ctx.exit(1)
                elif e.http_status != 404:
                    current_app.logger.error(
                        "create_transfer_endpoint: error checking if "
                        "endpoint already exists: %s",
                        str(e),
                    )
                    self.ctx.exit(1)

        if not hasattr(endpoint_model, "endpoint_id"):
            try:
                current_app.logger.debug(
                    "download_existing_configuration: initializing endpoint"
                )
                endpoint_model.fingerprint = {"GCS": __version_fingerprint__}
                endpoint_model.put(**self.endpoint)
                current_app.logger.debug(
                    "download_existing_configuration: saving endpoint"
                )
                endpoint_model = current_app.appsync_client.update_object(
                    endpoint_model
                )
                current_app.logger.debug(
                    "download_existing_configuration: add to session"
                )
                db.session.add(endpoint_model)
                current_app.logger.debug(
                    "download_existing_configuration: commit to database"
                )
                db.session.commit()
            except DataKeyError as e:
                current_app.logger.error(
                    "graphql operation cannot be processed: %s", str(e)
                )
                self.ctx.exit(1)
        else:
            current_app.logger.debug(
                "download_existing_configuration: updating endpoint"
            )

            res = current_app.transfer_client.get_endpoint(endpoint_model.endpoint_id)
            if "subscription_id" in res.data:
                self.endpoint["subscription_id"] = res.data["subscription_id"]

            endpoint_model.put(endpoint_id=endpoint_model.endpoint_id, **self.endpoint)

            # record endpoint document in appsync
            current_app.appsync_client.update_object(endpoint_model)
            current_app.logger.debug("download_existing_configuration: saving endpoint")
            current_app.logger.debug("download_existing_configuration: add to session")
            db.session.add(endpoint_model)
            current_app.logger.debug(
                "download_existing_configuration: commit to database"
            )
            db.session.commit()

        current_app.config["HOST_ENDPOINT_ID"] = endpoint_model.endpoint_id
        self.endpoint_model = endpoint_model
        current_app.logger.info("create_transfer_endpoint: start")

    def check_transfer_subscription(self):
        """
        Check subscription status
        """
        current_app.logger.info("check_transfer_subscription: start")

        ep_doc = current_app.transfer_client.get_endpoint(
            self.endpoint_model.endpoint_id
        )

        subscription_id = ep_doc.get("subscription_id")

        if subscription_id:
            # If the endpoint already has a subscription, update the model
            self.endpoint_model.patch(
                DATA_TYPE="endpoint#1.0.0",
                subscription_id=subscription_id,
            )
            current_app.appsync_client.update_object(self.endpoint_model)
            db.session.add(self.endpoint_model)
            db.session.commit()
            if self.subscription_id and self.subscription_id != subscription_id:
                current_app.logger.warn(
                    "check_transfer_subscription: subscription_id=%s " "requested=%s",
                    subscription_id,
                    self.subscription_id,
                )
                click.echo(
                    "Endpoint subscription_id does not match requested value", err=True
                )
            self.subscription_set = True
        current_app.logger.info("check_transfer_subscription: end")

    def create_roles(self) -> None:
        """
        Create owner role
        """
        current_app.logger.info("create_roles: start")
        role = Role.query.filter(
            ~Role.deleted,  # type: ignore
            Role.role == RoleType.endpoint_owner,
        ).one_or_none()

        principal = "urn:globus:auth:identity:" + self.owner_id
        if role:
            if role.identity_id != self.owner_id:
                current_app.logger.error(
                    "Owner role already exists for %s", role.identity_id
                )
                old_admin = current_app.auth_client.get_identities(
                    ids=[role.identity_id]
                ).data.get("identities", [])
                if len(old_admin) == 1:
                    old_user = old_admin[0]["username"]
                    click.echo("Owner role exists for {}".format(old_user))
                else:
                    click.echo("Owner role exists for {}".format(role.identity_id))

                if not click.confirm("Replace owner role?"):
                    self.ctx.exit(1)

                role.put_owner_role(self.owner_id)
                try:
                    role = current_app.appsync_client.update_object(role)
                    db.session.add(role)
                    db.session.commit()
                except (globus_sdk.exc.GlobusAPIError, errors.BadGatewayError) as e:
                    current_app.logger.error(
                        "Unable to update owner role: {}".format(str(e))
                    )
                    self.ctx.exit(1)
            else:
                current_app.logger.info("create_roles: reusing_existing_role")
        else:
            try:
                role = Role(
                    doc=dict(DATA_TYPE="role#1.0.0", principal=principal, role="owner"),
                    endpoint=self.endpoint_model,
                )
                role.primary_role = True
                role = current_app.appsync_client.create_object(role)
                db.session.add(role)
                db.session.commit()
            except (globus_sdk.exc.GlobusAPIError, errors.BadGatewayError) as e:
                current_app.logger.info(
                    "Unable to create owner role: {}".format(str(e))
                )

        if self.subscription_set:
            admin_role = Role.query.filter(
                Role.role == RoleType.endpoint_administrator,
                Role.principal == principal,
            ).one_or_none()
            if not admin_role:
                admin_role = Role(
                    doc=dict(
                        DATA_TYPE="role#1.0.0",
                        principal=principal,
                        role="administrator",
                    ),
                    endpoint=self.endpoint_model,
                    primary_role=True,
                )
                admin_role = current_app.appsync_client.create_object(admin_role)
                db.session.add(admin_role)
                db.session.commit()
        current_app.logger.info("create_roles: end")

    def pre_register_dns(self):
        """
        DNS registration (may take a few minutes)
        """
        pass

    def register_dns(self):
        """
        DNS registration
        """
        current_app.logger.info("register_dns: start")
        domain = GlobusDnsDomain.query.one_or_none()

        if not domain:
            domain = GlobusDnsDomain()
            current_app.appsync_client.create_object(domain)
            db.session.add(domain)
            db.session.commit()
        self.domain = domain.domain_name
        current_app.logger.info("register_dns: end")

    def create_acme_client(self):
        """
        Create Let's Encrypt account
        """
        current_app.logger.info("create_acme_account: start")
        acme_account = AcmeAccount.query.one_or_none()
        if not acme_account:
            acme_account = AcmeAccount()
            try:
                acme_account = current_app.appsync_client.create_object(acme_account)
            except GraphQLError as e:
                if len(e.args) > 1:
                    acme_account = AcmeAccount(appsync_data=e.args[1])
                else:
                    current_app.logger.error("Creating acme account: {}".format(e))
                    self.ctx.exit(1)

            db.session.add(acme_account)
            db.session.commit()
        self.acme_account = acme_account
        current_app.logger.info("create_acme_account: end")

    def pre_provision_certificate(self):
        """
        Get Certificate (may take a few minutes)
        """

    def provision_certificate(self):
        """
        Get Certificate
        """
        current_app.logger.info("provision_certificate: start")
        certificate = LetsEncryptCertificate.query.one_or_none()
        if not certificate:
            acme_account = AcmeAccount.query.one()
            acme_account.acquire_lease()

            certificate = LetsEncryptCertificate.create_certificate(self.domain)
            certificate = current_app.appsync_client.create_object(certificate)
            acme_account.release_lease()
            db.session.add(certificate)
            db.session.commit()
        current_app.logger.info("provision_certificate: end")

    def register_server(self):
        """
        Register with Globus
        """
        current_app.logger.info("register_server: start")
        self.endpoint_model.update_endpoint_server()
        current_app.logger.info("register_server: end")

    def set_owner(self):
        """
        Set endpoint advertised owner
        """
        assert self.transfer_token

        try:
            self.endpoint_model.set_owner_string(
                owner_id=self.owner_id, token=self.transfer_token["access_token"]
            )
        except Exception:
            click.echo(
                "\n".join(
                    textwrap.wrap(
                        "Unable to set the advertised owner id. Have the "
                        "endpoint owner or an endpoint administrator "
                        "log in to the endpoint and use the command "
                        "globus-connect-server endpoint set-owner-string "
                        "to set this."
                    )
                )
            )
